"""Configurations for Ralph."""

import io
from enum import Enum
from pathlib import Path
from typing import List, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

try:
    from click import get_app_dir
except ImportError:
    # If we use Ralph as a library and Click is not installed, we consider the
    # application directory to be the current directory. For non-CLI usage, it
    # has no consequences.
    from unittest.mock import Mock

    get_app_dir = Mock(return_value=".")
from pydantic import AnyHttpUrl, AnyUrl, BaseModel, BaseSettings, Extra, Field

from .utils import import_string

MODEL_PATH_SEPARATOR = "__"


class BaseSettingsConfig:
    """Pydantic model for BaseSettings Configuration."""

    case_sensitive = True
    env_nested_delimiter = "__"
    env_prefix = "RALPH_"


class CoreSettings(BaseSettings):
    """Pydantice model for Ralph's core settings."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

    APP_DIR: Path = get_app_dir("ralph")
    LOCALE_ENCODING: str = getattr(io, "LOCALE_ENCODING", "utf8")


core_settings = CoreSettings()


class CommaSeparatedTuple(str):
    """Pydantic field type validating comma separated strings or tuples."""

    @classmethod
    def __get_validators__(cls):  # noqa: D105
        def validate(value: Union[str, Tuple[str]]) -> Tuple[str]:
            """Checks whether the value is a comma separated string or a tuple."""
            if isinstance(value, tuple):
                return value

            if isinstance(value, str):
                return tuple(value.split(","))

            raise TypeError("Invalid comma separated list")

        yield validate


class InstantiableSettingsItem(BaseModel):
    """Pydantic model for a settings configuration item that can be instantiated."""

    class Config:  # pylint: disable=missing-class-docstring # noqa: D106
        underscore_attrs_are_private = True

    _class_path: str = None

    def get_instance(self, **init_parameters):
        """Returns an instance of the settings item class using it's `_class_path`."""
        return import_string(self._class_path)(**init_parameters)


# Active database backend Settings.


class ClientOptions(BaseModel):
    """Pydantic model for additionnal client options."""

    class Config:  # pylint: disable=missing-class-docstring # noqa: D106
        extra = Extra.forbid


class ClickhouseClientOptions(ClientOptions):
    """Pydantic model for `clickhouse` client options."""

    date_time_input_format: str = "best_effort"
    allow_experimental_object_type: Literal[0, 1] = None


class ESClientOptions(ClientOptions):
    """Pydantic model for Elasticsearch additionnal client options."""

    ca_certs: Path = None
    verify_certs: bool = None


class ClickhouseDatabaseBackendSettings(InstantiableSettingsItem):
    """Pydantic model for ClickHouse database backend configuration settings."""

    _class_path: str = "ralph.backends.database.clickhouse.ClickHouseDatabase"

    HOST: str = "localhost"
    PORT: int = 8123
    DATABASE: str = "xapi"
    EVENT_TABLE_NAME: str = "xapi_events_all"
    USERNAME: str = None
    PASSWORD: str = None
    CLIENT_OPTIONS: ClickhouseClientOptions = None


class MongoClientOptions(ClientOptions):
    """Pydantic model for MongoDB additionnal client options."""

    document_class: str = None
    tz_aware: bool = None


class ESDatabaseBackendSettings(InstantiableSettingsItem):
    """Pydantic modelf for Elasticsearch database backend configuration settings."""

    _class_path: str = "ralph.backends.database.es.ESDatabase"

    HOSTS: CommaSeparatedTuple = ("http://localhost:9200",)
    INDEX: str = "statements"
    CLIENT_OPTIONS: ESClientOptions = ESClientOptions()
    OP_TYPE: Literal["index", "create", "delete", "update"] = "index"


class MongoDatabaseBackendSettings(InstantiableSettingsItem):
    """Pydantic model for Mongo database backend configuration settings."""

    _class_path: str = "ralph.backends.database.mongo.MongoDatabase"

    CONNECTION_URI: str = "mongodb://localhost:27017/"
    DATABASE: str = "statements"
    COLLECTION: str = "marsha"
    CLIENT_OPTIONS: MongoClientOptions = MongoClientOptions()


class DatabaseBackendSettings(BaseModel):
    """Pydantic model for database backend configuration settings."""

    ES: ESDatabaseBackendSettings = ESDatabaseBackendSettings()
    MONGO: MongoDatabaseBackendSettings = MongoDatabaseBackendSettings()
    CLICKHOUSE: ClickhouseDatabaseBackendSettings = ClickhouseDatabaseBackendSettings()


# Active HTTP backend Settings.


class HeadersParameters(BaseModel):
    """Pydantic model for headers parameters."""

    class Config:  # pylint: disable=missing-class-docstring # noqa: D106
        extra = Extra.allow


class LRSHeaders(HeadersParameters):
    """Pydantic model for LRS headers."""

    X_EXPERIENCE_API_VERSION: str = "1.0.3"
    CONTENT_TYPE: str = "application/json"


class LRSHTTPBackendSettings(InstantiableSettingsItem):
    """Pydantic model for LRS HTTP backend configuration settings."""

    _class_path: str = "ralph.backends.http.lrs.LRSHTTP"

    BASE_URL: AnyHttpUrl = Field("http://0.0.0.0:8100")
    USERNAME: str = "ralph"
    PASSWORD: str = "secret"
    HEADERS: LRSHeaders = LRSHeaders()
    STATUS_ENDPOINT: str = "/__heartbeat__"
    STATEMENTS_ENDPOINT: str = "/xAPI/statements"


class HTTPBackendSettings(BaseModel):
    """Pydantic model for HTTP backend configuration settings."""

    LRS: LRSHTTPBackendSettings = LRSHTTPBackendSettings()


# Active storage backend Settings.


class FSStorageBackendSettings(InstantiableSettingsItem):
    """Pydantic model for FileSystem storage backend configuration settings."""

    _class_path: str = "ralph.backends.storage.fs.FSStorage"

    PATH: str = str(core_settings.APP_DIR / "archives")


class LDPStorageBackendSettings(InstantiableSettingsItem):
    """Pydantic model for LDP storage backend configuration settings."""

    _class_path: str = "ralph.backends.storage.ldp.LDPStorage"

    ENDPOINT: str = None
    APPLICATION_KEY: str = None
    APPLICATION_SECRET: str = None
    CONSUMER_KEY: str = None
    SERVICE_NAME: str = None
    STREAM_ID: str = None


class SWIFTStorageBackendSettings(InstantiableSettingsItem):
    """Pydantic model for SWIFT storage backend configuration settings."""

    _class_path: str = "ralph.backends.storage.swift.SwiftStorage"

    OS_TENANT_ID: str = None
    OS_TENANT_NAME: str = None
    OS_USERNAME: str = None
    OS_PASSWORD: str = None
    OS_REGION_NAME: str = None
    OS_STORAGE_URL: str = None
    OS_USER_DOMAIN_NAME: str = "Default"
    OS_PROJECT_DOMAIN_NAME: str = "Default"
    OS_AUTH_URL: str = "https://auth.cloud.ovh.net/"
    OS_IDENTITY_API_VERSION: str = "3"


class S3StorageBackendSettings(InstantiableSettingsItem):
    """Represents the S3 storage backend configuration settings."""

    _class_path: str = "ralph.backends.storage.s3.S3Storage"

    ACCESS_KEY_ID: str = None
    SECRET_ACCESS_KEY: str = None
    SESSION_TOKEN: str = None
    DEFAULT_REGION: str = None
    BUCKET_NAME: str = None
    ENDPOINT_URL: str = None


class StorageBackendSettings(BaseModel):
    """Pydantic model for storage backend configuration settings."""

    LDP: LDPStorageBackendSettings = LDPStorageBackendSettings()
    FS: FSStorageBackendSettings = FSStorageBackendSettings()
    SWIFT: SWIFTStorageBackendSettings = SWIFTStorageBackendSettings()
    S3: S3StorageBackendSettings = S3StorageBackendSettings()


# Active storage backend Settings.


class WSStreamBackendSettings(InstantiableSettingsItem):
    """Pydantic model for Websocket stream backend configuration settings."""

    _class_path: str = "ralph.backends.stream.ws.WSStream"

    URI: str = None


class StreamBackendSettings(BaseModel):
    """Pydantic model for stream backend configuration settings."""

    WS: WSStreamBackendSettings = WSStreamBackendSettings()


# Active backend Settings.


class BackendSettings(BaseModel):
    """Pydantic model for backends configuration settings."""

    DATABASE: DatabaseBackendSettings = DatabaseBackendSettings()
    HTTP: HTTPBackendSettings = HTTPBackendSettings()
    STORAGE: StorageBackendSettings = StorageBackendSettings()
    STREAM: StreamBackendSettings = StreamBackendSettings()


# Active parser Settings.


class ESParserSettings(InstantiableSettingsItem):
    """Pydantic model for Elastisearch parser configuration settings."""

    _class_path: str = "ralph.parsers.ElasticSearchParser"


class GELFParserSettings(InstantiableSettingsItem):
    """Pydantic model for GELF parser configuration settings."""

    _class_path: str = "ralph.parsers.GELFParser"


class ParserSettings(BaseModel):
    """Pydantic model for parsers configuration settings."""

    GELF: GELFParserSettings = GELFParserSettings()
    ES: ESParserSettings = ESParserSettings()


class XapiForwardingConfigurationSettings(BaseModel):
    """Pydantic model for xAPI forwarding configuration item."""

    class Config:  # pylint: disable=missing-class-docstring # noqa: D106
        min_anystr_length = 1

    url: AnyUrl
    is_active: bool
    basic_username: str
    basic_password: str
    max_retries: int
    timeout: float


class Settings(BaseSettings):
    """Pydantic model for Ralph's global environment & configuration settings."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_file = ".env"
        env_file_encoding = core_settings.LOCALE_ENCODING

    class AuthBackends(Enum):
        """Enum of the authentication backends."""

        BASIC = "basic"
        OIDC = "oidc"

    _CORE: CoreSettings = core_settings
    AUTH_FILE: Path = _CORE.APP_DIR / "auth.json"
    AUTH_CACHE_MAX_SIZE = 100
    AUTH_CACHE_TTL = 3600
    BACKENDS: BackendSettings = BackendSettings()
    CONVERTER_EDX_XAPI_UUID_NAMESPACE: str = None
    DEFAULT_BACKEND_CHUNK_SIZE: int = 500
    EXECUTION_ENVIRONMENT: str = "development"
    HISTORY_FILE: Path = _CORE.APP_DIR / "history.json"
    LOGGING: dict = {
        "version": 1,
        "propagate": True,
        "formatters": {
            "ralph": {
                "format": "%(asctime)-23s %(levelname)-8s %(name)-8s %(message)s"
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "stream": "ext://sys.stderr",
                "formatter": "ralph",
            },
        },
        "loggers": {
            "ralph": {
                "handlers": ["console"],
                "level": "INFO",
            },
            "swiftclient": {
                "handlers": ["console"],
                "level": "ERROR",
            },
            "uvicorn": {
                "handlers": ["console"],
                "level": "INFO",
            },
        },
    }
    PARSERS: ParserSettings = ParserSettings()
    RUNSERVER_AUTH_BACKEND: AuthBackends = AuthBackends.BASIC
    RUNSERVER_AUTH_OIDC_AUDIENCE: str = None
    RUNSERVER_AUTH_OIDC_ISSUER_URI: AnyHttpUrl = None
    RUNSERVER_BACKEND: Literal["clickhouse", "es", "mongo"] = "es"
    RUNSERVER_HOST: str = "0.0.0.0"  # nosec
    RUNSERVER_MAX_SEARCH_HITS_COUNT: int = 100
    RUNSERVER_POINT_IN_TIME_KEEP_ALIVE: str = "1m"
    RUNSERVER_PORT: int = 8100
    SENTRY_CLI_TRACES_SAMPLE_RATE: float = 1.0
    SENTRY_DSN: str = None
    SENTRY_IGNORE_HEALTH_CHECKS: bool = False
    SENTRY_LRS_TRACES_SAMPLE_RATE: float = 1.0
    XAPI_FORWARDINGS: List[XapiForwardingConfigurationSettings] = []

    @property
    def APP_DIR(self) -> Path:  # pylint: disable=invalid-name
        """Returns the path to Ralph's configuration directory."""
        return self._CORE.APP_DIR

    @property
    def LOCALE_ENCODING(self) -> str:  # pylint: disable=invalid-name
        """Returns Ralph's default locale encoding."""
        return self._CORE.LOCALE_ENCODING


settings = Settings()



"""Websocket stream backend for Ralph."""

import asyncio
import logging
from typing import BinaryIO

import websockets

from ralph.conf import settings

from .base import BaseStream

logger = logging.getLogger(__name__)


class WSStream(BaseStream):
    """Websocket stream backend."""

    name = "ws"

    def __init__(self, uri: str = settings.BACKENDS.STREAM.WS.URI):
        """Instantiates the websocket client.

        Args:
            uri (str): The URI to connect to.
        """
        self.uri = uri

    def stream(self, target: BinaryIO):
        """Streams websocket content to target."""
        # pylint: disable=no-member

        logger.debug("Streaming from websocket uri: %s", self.uri)

        async def _stream():
            async with websockets.connect(self.uri) as websocket:
                while event := await websocket.recv():
                    target.write(bytes(f"{event}" + "\n", encoding="utf-8"))

        asyncio.get_event_loop().run_until_complete(_stream())



"""Base Stream backend for Ralph."""

from abc import ABC, abstractmethod
from typing import BinaryIO


class BaseStream(ABC):
    """Base stream backend interface."""

    name = "base"

    @abstractmethod
    def stream(self, target: BinaryIO):
        """Reads records and streams them to target."""



"""Base storage backend for Ralph."""

from abc import ABC, abstractmethod
from typing import Iterable


class BaseStorage(ABC):
    """Base storage backend interface."""

    name = "base"

    @abstractmethod
    def list(self, details=False, new=False):
        """Lists files in the storage backend."""

    @abstractmethod
    def url(self, name):
        """Gets `name` file absolute URL."""

    @abstractmethod
    def read(self, name, chunk_size: int = 4096):
        """Reads `name` file and yields its content by chunks of a given size."""

    @abstractmethod
    def write(self, stream: Iterable, name, overwrite=False):
        """Writes content to the `name` target."""



"""`Video` activity types definitions."""

from typing import Dict, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition
from ...constants import LANG_EN_US_DISPLAY

# Video


class VideoActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for video `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/video/activity-type/video`.
    """

    type: Literal[
        "https://w3id.org/xapi/video/activity-type/video"
    ] = "https://w3id.org/xapi/video/activity-type/video"


class VideoActivity(BaseXapiActivity):
    """Pydantic model for video `Activity` type.

    WARNING: Contains an optional name property, this is not a violation of
    conformity but goes against xAPI specification recommendations.

    Attributes:
        name (dict): Consists of the dictionary `{"en-US": <name of the video>}`.
        definition (dict): See VideoActivityDefinition.
    """

    name: Optional[Dict[Literal[LANG_EN_US_DISPLAY], str]]
    definition: VideoActivityDefinition = VideoActivityDefinition()



"""Problem interaction events model event fields definitions."""

from datetime import datetime
from typing import Dict, List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from pydantic import constr

from ...base import AbstractBaseEventField, BaseModelWithConfig


class QueueState(BaseModelWithConfig):
    """Pydantic model for problem interaction `event`.`correct_map`.`queuestate` field.

    Attributes:
        key (str): Consists of a secret string.
        time (str): Consists of a string dump of a DateTime object in the format
            '%Y%m%d%H%M%S'.
    """

    key: str
    time: datetime


class CorrectMap(BaseModelWithConfig):
    """Pydantic model for problem interaction `event`.`correct_map` field.

    Attributes:
        answervariable (str): Consists of the variable chosen in answer in the case of
            optionresponse provided with variables.
        correctness (str): Consists either of the `correct` or `incorrect` value.
        hint (str): Consists of optional hint.
        hint_mode (str): Consists either of the value `on_request` or `always` value.
        msg (str): Consists of extra message response.
        npoints (int): Consists of awarded points.
        queuestate (json): see QueueStateField.
    """

    answervariable: Union[Literal[None], None, str]
    correctness: Union[Literal["correct"], Literal["incorrect"]]
    hint: Optional[str]
    hintmode: Optional[Union[Literal["on_request"], Literal["always"]]]
    msg: str
    npoints: Optional[int]
    queuestate: Optional[QueueState]


class State(BaseModelWithConfig):
    """Pydantic modelfor problem interaction `event`.`state` field.

    Attributes:
        correct_map (dict): see CorrectMapSubFields.
        done (bool): `True` if the problem is answered, else `False`.
        input_state (dict): Consists of the state field given before answering.
        seed (int): Consists of the seed element for the current state.
        student_answers (dict): Consists of the answer(s) given by the user.
    """

    correct_map: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),  # noqa : F722
        CorrectMap,
    ]
    done: Optional[bool]
    input_state: dict
    seed: int
    student_answers: dict


class SubmissionAnswerField(BaseModelWithConfig):
    """Pydantic model for `problem_check`.`event`.`submission` field.

    Attributes:
        answer (str, list): Consists of the answer string or a list of the answer
            strings if multiple choices are allorwed.
        correct (bool): `True` if the `answer` value is correct, else `False`.
        input_type (str): Consists of the type of value that the student supplies for
            the `response_type`.
        question (str): Consists of the question text.
        response_type (str): Consists of the type of problem.
        variant (str): Consists of the unique ID of the variant that was presented to
            this user.
    """

    answer: Union[str, List[str]]
    correct: bool
    input_type: str
    question: str
    response_type: str
    variant: str


class EdxProblemHintDemandhintDisplayedEventField(AbstractBaseEventField):
    """Pydantic model for `edx.problem.hint.demandhint_displayed`.`event` field.

    Attributes:
        hint_index (int): Consists of the identifier for the hint that was displayed to
            the user.
        hint_len (int): Consists of the total number of hints defined for this problem.
        hint_text (str): Consists of the text of the hint that was displayed to the
            user.
        module_id (str): Consists of the identifier for the problem component for which
            the user requested the hint.
    """

    hint_index: int
    hint_len: int
    hint_text: str
    module_id: str


class EdxProblemHintFeedbackDisplayedEventField(AbstractBaseEventField):
    """Pydantic model for `edx.problem.hint.feedback_displayed`.`event` field.

    Attributes:
        choice_all (list): Lists all of the answer choices for problems with multiple
            possible answers defined.
        correctness (bool): `True` if the `student_answer` value is correct, else
            `False`.
        hint_label (str): Consists of the feedback message given for the answer
            correctness.
        hints (list): Consists of a text member field with the given feedback string.
        module_id (str): Consists of the identifier for the problem component for which
            the user received the feedback.
        problem_part_id (str): Consists of the specific problem for which the user
            received feedback.
        question_type (str): Consists of the XML tag that identifies the problem type.
        student_answer (list): Consists of the answer value(s) selected or supplied by
            the user.
        trigger_type (str): Identifies the type of feedback obtained by the
            `student_answer` response. Consists either of `single` or `compound` value.
    """

    choice_all: Optional[List[str]]
    correctness: bool
    hint_label: str
    hints: List[dict]
    module_id: str
    problem_part_id: str
    question_type: Union[
        Literal["stringresponse"],
        Literal["choiceresponse"],
        Literal["multiplechoiceresponse"],
        Literal["numericalresponse"],
        Literal["optionresponse"],
    ]
    student_answer: List[str]
    trigger_type: Union[Literal["single"], Literal["compound"]]


class ProblemCheckEventField(AbstractBaseEventField):
    """Pydantic model for `problem_check`.`event` field.

    Attributes:
        answers (dict): Consists of a dictionary of problem ID and the corresponding
            internal answer identifier for each problem.
        attempts (int): Consists of the number of times the user attempted to answer
            the problem.
        correct_map (dict): Consists of the evaluation data for each answer.
        grade (int): Consists of the current grade value.
        max_grade (int): Consists of the maximum possible grade value.
        problem_id (str): Consists of the ID of the problem that was checked.
        state (json): Consists of the current problem state.
        submission (dict): Consists of a dictionnary of data about the given answer.
        success (str): Consists of either the `correct` or `incorrect` value.
    """

    answers: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),  # noqa : F722
        Union[List[str], str],
    ]
    attempts: int
    correct_map: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),  # noqa : F722
        CorrectMap,
    ]
    grade: int
    max_grade: int
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State
    submission: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),  # noqa : F722
        SubmissionAnswerField,
    ]
    success: Union[Literal["correct"], Literal["incorrect"]]


class ProblemCheckFailEventField(AbstractBaseEventField):
    """Pydantic model for `problem_check_fail`.`event` field.

    Attributes:
        answers (dict): Consists of a dictionary of problem ID and the internal answer
            identifier for each problem.
        failure (str): Consists either of the `closed` or `unreset` value.
        problem_id (str): Consists of the ID of the problem that was checked.
        state (dict): Consists of the current problem state.
    """

    answers: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),  # noqa : F722
        Union[List[str], str],
    ]
    failure: Union[Literal["closed"], Literal["unreset"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State


class ProblemRescoreEventField(AbstractBaseEventField):
    """Pydantic model for `problem_rescore`.`event` field.

    Attributes:
        attempts (int): Consists of the number of attempts of rescoring.
        correct_map (json): see CorrectMapSubFields.
        new_score (int): Consists of the new score obtained after rescoring.
        new_total (int): Consists of the new total summed after rescoring.
        orig_score (int): Consists of the original scored before rescoring.
        problem_id (str): Consists of the ID of the problem being rescored.
        state (json): see StateField.
        success (str): Consists either of the `correct` or `incorrect` value.
    """

    attempts: int
    correct_map: CorrectMap
    new_score: int
    new_total: int
    orig_score: int
    orig_total: int
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State
    success: Union[Literal["correct"], Literal["incorrect"]]


class ProblemRescoreFailEventField(AbstractBaseEventField):
    """Pydantic model for `problem_rescore_fail`.`event` field.

    Attributes:
        failure (str): Consists either of the `closed` or `unreset` value.
        problem_id (str): Consists of the ID of the problem being checked.
        state (json): see StateField.
    """

    failure: Union[Literal["closed"], Literal["unreset"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State


class UIProblemResetEventField(AbstractBaseEventField):
    """Pydantic model for `problem_reset`.`event` field.

    Attributes:
        answers (str, list): Consists of the answer string or a list of the answer
            strings if multiple choices are allowed.
    """

    answers: Union[str, List[str]]


class UIProblemShowEventField(AbstractBaseEventField):
    """Pydantic model for `problem_show`.`event` field.

    Attributes:
        problem (str): Consists of the optional name value that the course creators
            supply or the system-generated hash code for the problem being shown.
    """

    problem: str


class ResetProblemEventField(AbstractBaseEventField):
    """Pydantic model for `reset_problem`.`event` field.

    Attributes:
        new_state (json): see StateField.
        old_state (json): see StateField.
        problem_id (str): Consists of the ID of the problem being reset.
    """

    new_state: State
    old_state: State
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )


class ResetProblemFailEventField(AbstractBaseEventField):
    """Pydantic model for `reset_problem_fail`.`event` field.

    Attributes:
        failure (str): Consists either of `closed` or `not_done` value.
        old_state (json): see StateField.
        problem_id (str): Consists of the ID of the problem being reset.
    """

    failure: Union[Literal["closed"], Literal["not_done"]]
    old_state: State
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )


class SaveProblemFailEventField(AbstractBaseEventField):
    """Pydantic model for `save_problem_fail`.`event` field.

    Attributes:
        answers (dict): Consists of a dict of the answer string or a list or a dict of
            the answer strings if multiple choices are allowed.
        failure (str): Consists either of `closed` or `done` value.
        problem_id (str): Consists of the ID of the problem being saved.
        state (json): see StateField.
    """

    answers: Dict[str, Union[int, str, list, dict]]
    failure: Union[Literal["closed"], Literal["done"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State


class SaveProblemSuccessEventField(AbstractBaseEventField):
    """Pydantic model for `save_problem_success`.`event` field.

    Attributes:
        answers (dict): Consists of a dict of the answer string or a list or a dict of
            the answer strings if multiple choices are allowed.
        problem_id (str): Consists of the ID of the problem being saved.
        state (json): see StateField.
    """

    answers: Dict[str, Union[int, str, list, dict]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )
    state: State


class ShowAnswerEventField(AbstractBaseEventField):
    """Pydantic model for `show_answer`.`event` field.

    Attributes:
        problem_id (str): Consists of the ID of the problem being shown.
    """

    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"  # noqa : F722
        r"type@problem\+block@[a-f0-9]{32}$"  # noqa : F722
    )



"""Virtual classroom xAPI events context fields definitions."""

from datetime import datetime
from typing import List, Optional, Union
from uuid import UUID

try:
    from typing import Literal  # pylint: disable = ungrouped-imports
except ImportError:
    from typing_extensions import Literal

from pydantic import Field

from ..base.contexts import BaseXapiContext, BaseXapiContextContextActivities
from ..base.unnested_objects import BaseXapiActivity
from ..concepts.activity_types.scorm_profile import ProfileActivity
from ..concepts.activity_types.virtual_classroom import VirtualClassroomActivity
from ..concepts.constants.cmi5_profile import CONTEXT_EXTENSION_SESSION_ID
from ..concepts.constants.tincan_vocabulary import CONTEXT_EXTENSION_PLANNED_DURATION
from ..config import BaseExtensionModelWithConfig


class VirtualClassroomContextActivitiesCategory(BaseXapiActivity):
    # noqa: D205, D415
    """Pydantic model for virtual classroom `context`.`contextActivities`.`category`
    property.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/virtual-classroom`.
        definition (dict): see ProfileActivity.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom"
    ] = "https://w3id.org/xapi/virtual-classroom"
    definition: ProfileActivity


class VirtualClassroomContextContextActivities(BaseXapiContextContextActivities):
    """Pydantic model for virtual classroom `context`.`contextActivities` property.

    Attributes:
        category (list): see VirtualClassroomContextActivitiesCategory.
    """

    category: Union[
        VirtualClassroomContextActivitiesCategory,
        List[VirtualClassroomContextActivitiesCategory],
    ]


class VirtualClassroomContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for virtual classroom base `context`.`extensions` property.

    Attributes:
        session_id (str): Consists of the ID of the active session.
    """

    session_id: str = Field(alias=CONTEXT_EXTENSION_SESSION_ID, default="")


class VirtualClassroomContext(BaseXapiContext):
    """Pydantic model for virtual classroom base `context` property.

    Attributes:
        registration (uuid): the registration that the Statement is associated with.
        extensions (dict): see VirtualClassroomExtensions.
    """

    contextActivities: VirtualClassroomContextContextActivities
    registration: UUID
    extensions: VirtualClassroomContextExtensions


# Initialized statement


class VirtualClassroomInitializedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom initialized `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomInitializedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom initialized `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomInitializedContextExtensions.
    """

    extensions: VirtualClassroomInitializedContextExtensions


# Joined statement


class VirtualClassroomJoinedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom initialized `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomJoinedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom joined `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomJoinedContextExtensions.
    """

    extensions: VirtualClassroomJoinedContextExtensions


# Terminated statement


class VirtualClassroomTerminatedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom terminated `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomTerminatedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom terminated `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomInitializedContextExtensions.
    """

    extensions: VirtualClassroomTerminatedContextExtensions


class VirtualClassroomStartedPollContextActivities(
    VirtualClassroomContextContextActivities
):
    # noqa: D205, D415
    """Pydantic model for virtual classroom started poll `context`.`contextActivities`
    property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomStartedPollContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom started poll `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (dict): see
            VirtualClassroomStartedPollContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomStartedPollContextActivities


# Answered poll statement


class VirtualClassroomAnsweredPollContextActivities(
    VirtualClassroomContextContextActivities
):
    # noqa: D205, D415
    """Pydantic model for virtual classroom answered poll `context`.`contextActivities`
    property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomAnsweredPollContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom answered poll `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (dict): see VirtualClassroomAnsweredPollContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomAnsweredPollContextActivities


# Posted public message statement


class VirtualClassroomPostedPublicMessageContextActivities(
    VirtualClassroomContextContextActivities
):
    # noqa: D205, D415
    """Pydantic model for virtual classroom posted public message
    `context`.`contextActivities` property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomPostedPublicMessageContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom posted public message `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (list): see
            VirtualClassroomPostedPublicMessageContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomPostedPublicMessageContextActivities



"""Ralph CLI entrypoint."""

import json
import logging
import re
import sys
from inspect import isclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List

import bcrypt

try:
    import click
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        "You need to install 'cli' optional dependencies to use the ralph "
        "command: pip install ralph-malph[cli]"
    ) from err
try:
    import uvicorn
except ModuleNotFoundError:
    # This error will be catched in the runserver command. We should be able to
    # use all commands except the runserver command when lrs optional
    # dependencies are not installed.
    pass
from click_option_group import optgroup
from pydantic import BaseModel

from ralph import __version__ as ralph_version
from ralph.conf import ClientOptions, CommaSeparatedTuple, HeadersParameters, settings
from ralph.exceptions import UnsupportedBackendException
from ralph.logger import configure_logging
from ralph.models.converter import Converter
from ralph.models.selector import ModelSelector
from ralph.models.validator import Validator
from ralph.utils import (
    get_backend_instance,
    get_backend_type,
    get_root_logger,
    import_string,
)

# cli module logger
logger = logging.getLogger(__name__)


class CommaSeparatedTupleParamType(click.ParamType):
    """Comma separated tuple parameter type."""

    name = "value1,value2,value3"

    def convert(self, value, param, ctx):
        """Splits the value by comma to return a tuple of values."""
        if isinstance(value, str):
            return tuple(value.split(","))

        if not isinstance(value, tuple):
            self.fail(
                "You should provide values separated by commas, e.g. foo,bar,baz",
                param,
                ctx,
            )

        return value


class CommaSeparatedKeyValueParamType(click.ParamType):
    """Comma separated key=value parameter type."""

    name = "key=value,key=value"

    def convert(self, value, param, ctx):
        """Splits the values by comma and equal sign.

        Returns a dictionary build with key/value pairs.
        """
        if isinstance(value, dict):
            return value

        # Parse options string
        try:
            options = dict(option.split("=") for option in value.split(","))
        except ValueError:
            self.fail(
                (
                    "You should provide key=value pairs separated by commas, "
                    "e.g. foo=bar,bar=2"
                ),
                param,
                ctx,
            )

        # Cast simple types
        for key, value_ in options.items():
            if value_ == "":
                options.update({key: None})
            elif re.match("^true$", value_, re.IGNORECASE):
                options.update({key: True})
            elif re.match("^false$", value_, re.IGNORECASE):
                options.update({key: False})
            elif re.match(r"^[0-9]+\.[0-9]+$", value_):
                options.update({key: float(value_)})
            elif re.match("^[0-9]+$", value_):
                options.update({key: int(value_)})

        return options


class ClientOptionsParamType(CommaSeparatedKeyValueParamType):
    """Comma separated key=value parameter type for client options."""

    def __init__(self, client_options_type):
        """Instantiates ClientOptionsParamType for a client_options_type.

        Args:
            client_options_type (any): Pydantic model used for client options.
        """
        self.client_options_type = client_options_type

    def convert(self, value, param, ctx):
        """Splits the values by comma and equal sign.

        Returns an instance of client_options_type build with key/value pairs.
        """
        if isinstance(value, self.client_options_type):
            return value

        return self.client_options_type(**super().convert(value, param, ctx))


class HeadersParametersParamType(CommaSeparatedKeyValueParamType):
    """Comma separated key=value parameter type for headers parameters."""

    def __init__(self, headers_parameters_type):
        """Instantiates HeadersParametersParamType for a headers_parameters_type.

        Args:
            headers_parameters_type (any): Pydantic model used for headers parameters.
        """
        self.headers_parameters_type = headers_parameters_type

    def convert(self, value, param, ctx):
        """Splits the values by comma and equal sign.

        Returns an instance of headers_parameters_type build with key/value pairs.
        """
        if isinstance(value, self.headers_parameters_type):
            return value

        return self.headers_parameters_type(**super().convert(value, param, ctx))


class JSONStringParamType(click.ParamType):
    """JSON string parameter type."""

    name = '\'{"key": "value", "key": "value"}\''

    def convert(self, value, param, ctx):
        """Load value as a json string and return a dict."""
        try:
            options = json.loads(value)
        except (json.JSONDecodeError, TypeError):
            self.fail(
                ("You should provide a valid JSON string as input"),
                param,
                ctx,
            )
        return options


@click.group(name="ralph")
@click.option(
    "-v",
    "--verbosity",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    metavar="LVL",
    required=False,
    help="Either CRITICAL, ERROR, WARNING, INFO (default) or DEBUG",
)
@click.version_option(version=ralph_version)
def cli(verbosity=None):
    """Ralph is a stream-based tool to play with your logs."""
    configure_logging()
    if verbosity is not None:
        level = getattr(logging, verbosity, None)
        get_root_logger().setLevel(level)
        for handler in get_root_logger().handlers:
            handler.setLevel(level)


def backends_options(name=None, backend_types: List[BaseModel] = None):
    """Backend-related options decorator for Ralph commands."""

    def wrapper(command):
        backend_names = []
        for backend_type in backend_types:
            for backend_name, backend in backend_type:
                backend_name = backend_name.lower()
                backend_names.append(backend_name)
                for field_name, field in backend:
                    field_type = backend.__fields__[field_name].type_
                    field_name = f"{backend_name}-{field_name}".replace("_", "-")
                    option = f"--{field_name}"
                    option_kwargs = {}
                    # If the field is a boolean, convert it to a flag option
                    if field_type is bool:
                        option = f"{option}/--no-{field_name}"
                        option_kwargs["is_flag"] = True
                    elif field_type is dict:
                        option_kwargs["type"] = CommaSeparatedKeyValueParamType()
                    elif field_type is CommaSeparatedTuple:
                        option_kwargs["type"] = CommaSeparatedTupleParamType()
                    elif isclass(field_type) and issubclass(field_type, ClientOptions):
                        option_kwargs["type"] = ClientOptionsParamType(field_type)
                    elif isclass(field_type) and issubclass(
                        field_type, HeadersParameters
                    ):
                        option_kwargs["type"] = HeadersParametersParamType(field_type)

                    command = optgroup.option(
                        option.lower(), default=field, **option_kwargs
                    )(command)

                command = (optgroup.group(f"{backend_name} backend"))(command)

        command = click.option(
            "-b",
            "--backend",
            type=click.Choice(backend_names),
            required=True,
            help="Backend",
        )(command)

        command = (cli.command(name=name or command.__name__))(command)
        return command

    return wrapper


@cli.command()
@click.option(
    "-u",
    "--username",
    type=str,
    required=True,
    help="The user for which we generate credentials.",
)
@click.password_option(
    "-p",
    "--password",
    type=str,
    required=True,
    help="The password to encrypt for this user. Will be prompted if missing.",
)
@click.option(
    "-s",
    "--scope",
    type=str,
    required=True,
    multiple=True,
    default=[],
    help="The user scope(s). This option can be provided multiple times.",
)
@click.option(
    "-w",
    "--write",
    is_flag=True,
    default=False,
    help="Write new credentials to the LRS authentication file.",
)
def auth(username, password, scope, write):
    """Generate credentials for LRS HTTP basic authentification."""
    logger.info("Will generate credentials for user: %s", username)

    # Import required Pydantic models dynamically so that we don't create a
    # direct dependency between the CLI and the LRS
    # pylint: disable=invalid-name
    ServerUsersCredentials = import_string(
        "ralph.api.auth.basic.ServerUsersCredentials"
    )
    UserCredentials = import_string("ralph.api.auth.basic.UserCredentials")

    credentials = UserCredentials(
        username=username,
        hash=bcrypt.hashpw(
            bytes(password, encoding=settings.LOCALE_ENCODING), bcrypt.gensalt()
        ).decode("ascii"),
        scopes=scope,
    )

    if write:
        logger.info("Will append new credentials to: %s", settings.AUTH_FILE)

        # Force Path object instantiation so that the file creation can be
        # faked in a test environment.
        auth_file = Path(settings.AUTH_FILE)
        # Create the authentication file if it does not exist
        auth_file.parent.mkdir(parents=True, exist_ok=True)
        auth_file.touch()

        users = ServerUsersCredentials.parse_obj([])
        # Parse credentials file if not empty
        if auth_file.stat().st_size:
            users = ServerUsersCredentials.parse_file(auth_file)
        users += ServerUsersCredentials.parse_obj(
            [
                credentials,
            ]
        )
        auth_file.write_text(users.json(indent=2), encoding=settings.LOCALE_ENCODING)
        logger.info("User %s has been added to: %s", username, settings.AUTH_FILE)
    else:
        click.echo(
            (
                f"Copy/paste the following credentials to your LRS authentication "
                f"file located in: {settings.AUTH_FILE}\n"
                f"{credentials.json(indent=2)}"
            )
        )


@cli.command()
@click.option(
    "-p",
    "--parser",
    type=click.Choice([parser.lower() for parser, _ in settings.PARSERS]),
    required=True,
    help="Container format parser used to extract events",
)
def extract(parser):
    """Extract input events from a container format using a dedicated parser."""
    logger.info("Extracting events using the %s parser", parser)

    parser = getattr(settings.PARSERS, parser.upper()).get_instance()

    for event in parser.parse(sys.stdin):
        click.echo(event)


@cli.command()
@click.option(
    "-f",
    "--format",
    "format_",
    type=click.Choice(["edx", "xapi"]),
    required=True,
    help="Input events format to validate",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue validating regardless of raised errors",
)
@click.option(
    "-F",
    "--fail-on-unknown",
    default=False,
    is_flag=True,
    help="Stop validating at first unknown event",
)
def validate(format_, ignore_errors, fail_on_unknown):
    """Validate input events of given format."""
    logger.info(
        "Validating %s events (ignore_errors=%s | fail-on-unknown=%s)",
        format_,
        ignore_errors,
        fail_on_unknown,
    )

    validator = Validator(ModelSelector(f"ralph.models.{format_}"))

    for event in validator.validate(sys.stdin, ignore_errors, fail_on_unknown):
        click.echo(event)


@cli.command()
@optgroup.group("From edX to xAPI converter options")
@optgroup.option(
    "-u",
    "--uuid-namespace",
    type=str,
    required=False,
    default=settings.CONVERTER_EDX_XAPI_UUID_NAMESPACE,
    help="The UUID namespace to use for the `ID` field generation",
)
@optgroup.option(
    "-p",
    "--platform-url",
    type=str,
    required=True,
    help="The `actor.account.homePage` to use in the xAPI statements",
)
@click.option(
    "-f",
    "--from",
    "from_",
    type=click.Choice(["edx"]),
    required=True,
    help="Input events format to convert",
)
@click.option(
    "-t",
    "--to",
    "to_",
    type=click.Choice(["xapi"]),
    required=True,
    help="Output events format",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue writing regardless of raised errors",
)
@click.option(
    "-F",
    "--fail-on-unknown",
    default=False,
    is_flag=True,
    help="Stop converting at first unknown event",
)
def convert(from_, to_, ignore_errors, fail_on_unknown, **conversion_set_kwargs):
    """Convert input events to a given format."""
    logger.info(
        "Converting %s events to %s format (ignore_errors=%s | fail-on-unknown=%s)",
        from_,
        to_,
        ignore_errors,
        fail_on_unknown,
    )
    logger.debug("Converter parameters: %s", conversion_set_kwargs)

    converter = Converter(
        model_selector=ModelSelector(f"ralph.models.{from_}"),
        module=f"ralph.models.{from_}.converters.{to_}",
        **conversion_set_kwargs,
    )

    for event in converter.convert(sys.stdin, ignore_errors, fail_on_unknown):
        click.echo(event)


@click.argument("archive", required=False)
@backends_options(backend_types=[backend for _, backend in settings.BACKENDS])
@click.option(
    "-c",
    "--chunk-size",
    type=int,
    default=settings.DEFAULT_BACKEND_CHUNK_SIZE,
    help="Get events by chunks of size #",
)
@click.option(
    "-t",
    "--target",
    type=str,
    default=None,
    help="Endpoint from which to fetch events (e.g. `/statements`)",
)
@click.option(
    "-q",
    "--query",
    type=JSONStringParamType(),
    default=None,
    help="Query object as a JSON string (database backends ONLY)",
)
def fetch(backend, archive, chunk_size, target, query, **options):
    """Fetch an archive or records from a configured backend."""
    logger.info(
        (
            "Fetching data from the configured %s backend "
            "(archive: %s | chunk size: %s | target: %s | query: %s)"
        ),
        backend,
        archive,
        chunk_size,
        target,
        query,
    )
    logger.debug("Backend parameters: %s", options)

    backend_type = get_backend_type(settings.BACKENDS, backend)
    backend = get_backend_instance(backend_type, backend, options)

    if backend_type == settings.BACKENDS.STORAGE:
        for data in backend.read(archive, chunk_size=chunk_size):
            click.echo(data, nl=False)
    elif backend_type == settings.BACKENDS.DATABASE:
        if query is not None:
            query = backend.query_model.parse_obj(query)
        for document in backend.get(query=query, chunk_size=chunk_size):
            click.echo(
                bytes(
                    json.dumps(document) if isinstance(document, dict) else document,
                    encoding="utf-8",
                )
            )
    elif backend_type == settings.BACKENDS.STREAM:
        backend.stream(sys.stdout.buffer)
    elif backend_type == settings.BACKENDS.HTTP:
        for statement in backend.read(target=target, chunk_size=chunk_size):
            click.echo(
                bytes(
                    json.dumps(statement) if isinstance(statement, dict) else statement,
                    encoding="utf-8",
                )
            )
    elif backend_type is None:
        msg = "Cannot find an implemented backend type for backend %s"
        logger.error(msg, backend)
        raise UnsupportedBackendException(msg, backend)


# pylint: disable=unnecessary-direct-lambda-call, too-many-arguments
@click.argument("archive", required=False)
@backends_options(
    backend_types=[
        settings.BACKENDS.DATABASE,
        settings.BACKENDS.STORAGE,
        settings.BACKENDS.HTTP,
    ]
)
@click.option(
    "-c",
    "--chunk-size",
    type=int,
    default=settings.DEFAULT_BACKEND_CHUNK_SIZE,
    help="Get events by chunks of size #",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    help="Overwrite existing archives or records",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue writing regardless of raised errors",
)
@click.option(
    "-t",
    "--target",
    type=str,
    default=None,
    help="Endpoint in which to push events (e.g. `statements`)",
)
def push(backend, archive, chunk_size, force, ignore_errors, target, **options):
    """Push an archive to a configured backend."""
    logger.info("Pushing archive %s to the configured %s backend", archive, backend)
    logger.debug("Backend parameters: %s", options)

    backend_type = get_backend_type(settings.BACKENDS, backend)
    backend = get_backend_instance(backend_type, backend, options)

    if backend_type == settings.BACKENDS.STORAGE:
        backend.write(sys.stdin.buffer, archive, overwrite=force)
    elif backend_type == settings.BACKENDS.DATABASE:
        backend.put(sys.stdin, chunk_size=chunk_size, ignore_errors=ignore_errors)
    elif backend_type == settings.BACKENDS.HTTP:
        backend.write(
            target=target,
            data=sys.stdin.buffer,
            chunk_size=chunk_size,
            ignore_errors=ignore_errors,
        )
    elif backend_type is None:
        msg = "Cannot find an implemented backend type for backend %s"
        logger.error(msg, backend)
        raise UnsupportedBackendException(msg, backend)


@backends_options(name="list", backend_types=[settings.BACKENDS.STORAGE])
@click.option(
    "-n/-a",
    "--new/--all",
    default=False,
    help="List not fetched (or all) archives",
)
@click.option(
    "-D/-I",
    "--details/--ids",
    default=False,
    help="Get archives detailed output (JSON)",
)
def list_(details, new, backend, **options):
    """List available archives from a configured storage backend."""
    logger.info("Listing archives for the configured %s backend", backend)
    logger.debug("Fetch details: %s", str(details))
    logger.debug("Backend parameters: %s", options)

    storage = get_backend_instance(settings.BACKENDS.STORAGE, backend, options)

    archives = storage.list(details=details, new=new)

    counter = 0
    for archive in archives:
        click.echo(json.dumps(archive) if details else archive)
        counter += 1

    if counter == 0:
        logger.warning("Configured %s backend contains no archive", backend)


@backends_options(name="runserver", backend_types=[settings.BACKENDS.DATABASE])
@click.option(
    "-h",
    "--host",
    type=str,
    required=False,
    default=settings.RUNSERVER_HOST,
    help="LRS server host name",
)
@click.option(
    "-p",
    "--port",
    type=int,
    required=False,
    default=settings.RUNSERVER_PORT,
    help="LRS server port",
)
def runserver(backend: str, host: str, port: int, **options):
    """Run the API server for the development environment.

    Starts uvicorn programmatically for convenience and documentation.
    """
    logger.info("Running API server on %s:%s with %s backend", host, port, backend)
    logger.info(
        (
            "Do not use runserver in production - start production servers "
            "through a process manager such as gunicorn/supervisor/circus."
        )
    )

    # The LRS server relies only on environment and configuration variables for its
    # configuration (not CLI arguments). Therefore, we convert the CLI arguments to
    # environment variables by creating a temporary environment file and passing it to
    # uvicorn.
    with NamedTemporaryFile(mode="w", encoding=settings.LOCALE_ENCODING) as env_file:
        env_file.write(f"RALPH_RUNSERVER_BACKEND={backend}\n")
        for key, value in options.items():
            if value is None:
                continue
            backend_name, field_name = key.split(sep="_", maxsplit=1)
            key = f"RALPH_BACKENDS__DATABASE__{backend_name}__{field_name}".upper()
            if isinstance(value, tuple):
                value = ",".join(value)
            if issubclass(type(value), ClientOptions):
                for key_dict, value_dict in value.dict().items():
                    if value_dict is None:
                        continue
                    key_dict = f"{key}__{key_dict}"
                    logger.debug(
                        "Setting environment variable %s to '%s'", key_dict, value_dict
                    )
                    env_file.write(f"{key_dict}={value_dict}\n")
                continue
            logger.debug("Setting environment variable %s to '%s'", key, value)
            env_file.write(f"{key}={value}\n")
        env_file.seek(0)
        try:
            uvicorn.run(
                "ralph.api:app",
                env_file=env_file.name,
                host=host,
                port=port,
                log_level="debug",
                reload=True,
            )
        except NameError as err:  # pylint: disable=redefined-outer-name
            raise ModuleNotFoundError(
                "You need to install 'lrs' optional dependencies to use the runserver "
                "command: pip install ralph-malph[lrs]"
            ) from err

    logger.info("Shutting down uvicorn server.")



"""Swift storage backend for Ralph."""

import logging
from functools import cached_property
from urllib.parse import urlparse

from swiftclient.service import SwiftService, SwiftUploadObject

from ralph.conf import settings
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

swift_settings = settings.BACKENDS.STORAGE.SWIFT
logger = logging.getLogger(__name__)


class SwiftStorage(
    HistoryMixin, BaseStorage
):  # pylint: disable=too-many-instance-attributes
    """OpenStack's Swift storage backend."""

    name = "swift"

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        os_tenant_id: str = swift_settings.OS_TENANT_ID,
        os_tenant_name: str = swift_settings.OS_TENANT_NAME,
        os_username: str = swift_settings.OS_USERNAME,
        os_password: str = swift_settings.OS_PASSWORD,
        os_region_name: str = swift_settings.OS_REGION_NAME,
        os_storage_url: str = swift_settings.OS_STORAGE_URL,
        os_user_domain_name: str = swift_settings.OS_USER_DOMAIN_NAME,
        os_project_domain_name: str = swift_settings.OS_PROJECT_DOMAIN_NAME,
        os_auth_url: str = swift_settings.OS_AUTH_URL,
        os_identity_api_version: str = swift_settings.OS_IDENTITY_API_VERSION,
    ):
        """Prepares the options for the SwiftService."""
        self.os_tenant_id = os_tenant_id
        self.os_tenant_name = os_tenant_name
        self.os_username = os_username
        self.os_password = os_password
        self.os_region_name = os_region_name
        self.os_user_domain_name = os_user_domain_name
        self.os_project_domain_name = os_project_domain_name
        self.os_auth_url = os_auth_url
        self.os_identity_api_version = os_identity_api_version
        self.container = urlparse(os_storage_url).path.rpartition("/")[-1]
        self.os_storage_url = os_storage_url
        if os_storage_url.endswith(f"/{self.container}"):
            self.os_storage_url = os_storage_url[: -len(f"/{self.container}")]

        with SwiftService(self.options) as swift:
            stats = swift.stat()
            if not stats["success"]:
                msg = "Unable to connect to the requested container: %s"
                logger.error(msg, stats["error"])
                raise BackendParameterException(msg % stats["error"])

    @cached_property
    def options(self):
        """Returns the required options for the SwiftService."""
        return {
            "os_auth_url": self.os_auth_url,
            "os_identity_api_version": self.os_identity_api_version,
            "os_password": self.os_password,
            "os_project_domain_name": self.os_project_domain_name,
            "os_region_name": self.os_region_name,
            "os_storage_url": self.os_storage_url,
            "os_tenant_id": self.os_tenant_id,
            "os_tenant_name": self.os_tenant_name,
            "os_username": self.os_username,
            "os_user_domain_name": self.os_user_domain_name,
        }

    def list(self, details=False, new=False):
        """Lists files in the storage backend."""
        archives_to_skip = set()
        if new:
            archives_to_skip = set(self.get_command_history(self.name, "fetch"))
        with SwiftService(self.options) as swift:
            for page in swift.list(self.container):
                if not page["success"]:
                    msg = "Failed to list container %s: %s"
                    logger.error(msg, page["container"], page["error"])
                    raise BackendException(msg % (page["container"], page["error"]))
                for archive in page["listing"]:
                    if new and archive["name"] in archives_to_skip:
                        continue
                    yield archive if details else archive["name"]

    def url(self, name):
        """Gets `name` file absolute URL."""
        # What's the purpose of this function ? Seems not used anywhere.
        return f"{self.options.get('os_storage_url')}/{name}"

    def read(self, name, chunk_size=None):
        """Reads `name` object and yields its content in chunks of (max) 2 ** 16.

        Why chunks of (max) 2 ** 16 ?
            Because SwiftService opens a file to stream the object into:
            See swiftclient.service.py:2082 open(filename, 'rb', DISK_BUFFER)
            Where filename = "/dev/stdout" and DISK_BUFFER = 2 ** 16
        """
        logger.debug("Getting archive: %s", name)

        with SwiftService(self.options) as swift:
            options = {"out_file": "-"}
            download = next(swift.download(self.container, [name], options), {})
        if "contents" not in download:
            msg = "Failed to download %s: %s"
            error = download.get("error", "swift.download did not yield")
            logger.error(msg, download.get("object", name), error)
            raise BackendException(msg % (download.get("object", name), error))
        size = 0
        for chunk in download["contents"]:
            logger.debug("Chunk %s", len(chunk))
            size += len(chunk)
            yield chunk

        # Archive fetched, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "size": size,
                "fetched_at": now(),
            }
        )

    def write(self, stream, name, overwrite=False):
        """Writes data from `stream` to the `name` target in chunks of (max) 2 ** 16."""
        if not overwrite and name in list(self.list()):
            msg = "%s already exists and overwrite is not allowed"
            logger.error(msg, name)
            raise FileExistsError(msg % name)

        logger.debug("Creating archive: %s", name)

        swift_object = SwiftUploadObject(stream, object_name=name)
        with SwiftService(self.options) as swift:
            for upload in swift.upload(self.container, [swift_object]):
                if not upload["success"]:
                    logger.error(upload["error"])
                    raise BackendException(upload["error"])

        # Archive written, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "push",
                "id": name,
                "pushed_at": now(),
            }
        )



"""Tests for Ralph utils."""

import pytest
from pydantic import BaseModel

from ralph import utils as ralph_utils
from ralph.conf import InstantiableSettingsItem, settings


def test_utils_import_string():
    """Tests import_string utility taken from Django utilities."""
    with pytest.raises(ImportError, match="foo doesn't look like a module path"):
        ralph_utils.import_string("foo")

    with pytest.raises(
        ImportError, match='Module "requests" does not define a "foo" attribute/class'
    ):
        ralph_utils.import_string("requests.foo")

    http_status = ralph_utils.import_string("http.HTTPStatus")
    assert http_status.OK == 200


def test_utils_get_backend_type():
    """Tests get_backend_type utility."""
    assert (
        ralph_utils.get_backend_type(settings.BACKENDS, "es")
        == settings.BACKENDS.DATABASE
    )
    assert (
        ralph_utils.get_backend_type(settings.BACKENDS, "ldp")
        == settings.BACKENDS.STORAGE
    )
    assert (
        ralph_utils.get_backend_type(settings.BACKENDS, "ws")
        == settings.BACKENDS.STREAM
    )
    assert ralph_utils.get_backend_type(settings.BACKENDS, "foo") is None


@pytest.mark.parametrize(
    "options,expected",
    [
        # Empty options should produce default result.
        ({}, {}),
        # Options not matching the backend name are ignored.
        ({"foo": "bar", "not_dummy_foo": "baz"}, {}),
        # One option matches the backend name and overrides the default.
        ({"dummy_foo": "bar", "not_dummy_foo": "baz"}, {"foo": "bar"}),
    ],
)
def test_utils_get_backend_instance(options, expected):
    """Tests get_backend_instance utility should return the expected result."""

    class DummyBackendSettings(InstantiableSettingsItem):
        """Represents a dummy backend setting."""

        foo: str = "foo"  # pylint: disable=disallowed-name

        def get_instance(self, **init_parameters):  # pylint: disable=no-self-use
            """Returns the init_parameters."""
            return init_parameters

    class TestBackendType(BaseModel):
        """A backend type including the DummyBackendSettings."""

        DUMMY: DummyBackendSettings = DummyBackendSettings()

    backend_instance = ralph_utils.get_backend_instance(
        TestBackendType(), "dummy", options
    )
    assert isinstance(backend_instance, dict)
    assert backend_instance == expected


@pytest.mark.parametrize("path,value", [(["foo", "bar"], "bar_value")])
def test_utils_get_dict_value_from_path_should_return_given_value(path, value):
    """Tests the get_dict_value_from_path function should return the value when it's
    present.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    assert ralph_utils.get_dict_value_from_path(dictionary, path) == value


@pytest.mark.parametrize(
    "path",
    [
        ["foo", "bar", "baz", "qux"],
        ["foo", "not_bar"],
        ["not_foo", "bar"],
        None,
    ],
)
def test_utils_get_dict_value_from_path_should_return_none_when_value_does_not_exists(
    path,
):
    """Tests the get_dict_value_from_path function should return None if the value is
    not found.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    assert ralph_utils.get_dict_value_from_path(dictionary, path) is None


def test_utils_set_dict_value_from_path_creating_new_fields():
    """Tests when the fields are not present, set_dict_value_from_path should add
    them.
    """
    dictionary = {}
    ralph_utils.set_dict_value_from_path(dictionary, ["foo", "bar"], "baz")
    assert dictionary == {"foo": {"bar": "baz"}}


def test_utils_set_dict_value_from_path_updating_fields():
    """Tests when the fields are present, set_dict_value_from_path should update
    them.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    ralph_utils.set_dict_value_from_path(dictionary, ["foo", "bar"], "baz")
    assert dictionary == {"foo": {"bar": "baz"}}



"""Elasticsearch database backend for Ralph."""

import json
import logging
from enum import Enum
from typing import Callable, Generator, List, Optional, TextIO

from elasticsearch import ApiError
from elasticsearch import ConnectionError as ESConnectionError
from elasticsearch import Elasticsearch
from elasticsearch.client import CatClient
from elasticsearch.helpers import BulkIndexError, scan, streaming_bulk

from ralph.conf import ESClientOptions, settings
from ralph.exceptions import BackendException, BackendParameterException

from .base import (
    BaseDatabase,
    BaseQuery,
    DatabaseStatus,
    StatementParameters,
    StatementQueryResult,
    enforce_query_checks,
)

es_settings = settings.BACKENDS.DATABASE.ES
logger = logging.getLogger(__name__)


class OpType(Enum):
    """Elasticsearch operation types."""

    INDEX = "index"
    CREATE = "create"
    DELETE = "delete"
    UPDATE = "update"


class ESQuery(BaseQuery):
    """Elasticsearch body query model."""

    query: Optional[dict]


class ESDatabase(BaseDatabase):
    """Elasticsearch database backend."""

    name = "es"
    query_model = ESQuery

    def __init__(
        self,
        hosts: list = es_settings.HOSTS,
        index: str = es_settings.INDEX,
        client_options: ESClientOptions = es_settings.CLIENT_OPTIONS,
        op_type: str = es_settings.OP_TYPE,
    ):
        """Instantiates the Elasticsearch client.

        Args:
            hosts (list): List of Elasticsearch nodes we should connect to.
            index (str): The Elasticsearch index name.
            client_options (dict): A dictionary of valid options for the
                Elasticsearch class initialization.
            op_type (str): The Elasticsearch operation type for every document sent to
                Elasticsearch (should be one of: index, create, delete, update).
        """
        self._hosts = hosts
        self.index = index

        self.client = Elasticsearch(self._hosts, **client_options.dict())
        if op_type not in [op.value for op in OpType]:
            raise BackendParameterException(
                f"{op_type} is not an allowed operation type"
            )
        self.op_type = op_type

    def status(self) -> DatabaseStatus:
        """Checks Elasticsearch cluster (connection) status."""
        # Check ES cluster connection
        try:
            self.client.info()
        except ESConnectionError:
            return DatabaseStatus.AWAY

        # Check cluster status
        if "green" not in CatClient(self.client).health():
            return DatabaseStatus.ERROR

        return DatabaseStatus.OK

    @enforce_query_checks
    def get(self, query: ESQuery = None, chunk_size: int = 500):
        """Gets index documents and yields them.

        The `query` dictionary should only contains kwargs compatible with the
        elasticsearch.helpers.scan function signature (API reference
        documentation:
        https://elasticsearch-py.readthedocs.io/en/latest/helpers.html#scan).
        """
        for document in scan(
            self.client, index=self.index, size=chunk_size, **query.dict()
        ):
            yield document

    def to_documents(
        self, stream: TextIO, get_id: Callable[[dict], str]
    ) -> Generator[dict, None, None]:
        """Converts `stream` lines to ES documents."""
        for line in stream:
            item = json.loads(line) if isinstance(line, str) else line
            action = {
                "_index": self.index,
                "_id": get_id(item),
                "_op_type": self.op_type,
            }
            if self.op_type == "update":
                action.update({"doc": item})
            elif self.op_type in ("create", "index"):
                action.update({"_source": item})
            yield action

    def put(
        self, stream: TextIO, chunk_size: int = 500, ignore_errors: bool = False
    ) -> int:
        """Writes documents from the `stream` to the instance index."""
        logger.debug(
            "Start writing to the %s index (chunk size: %d)", self.index, chunk_size
        )

        documents = 0
        try:
            for success, action in streaming_bulk(
                client=self.client,
                actions=self.to_documents(stream, lambda d: d.get("id", None)),
                chunk_size=chunk_size,
                raise_on_error=(not ignore_errors),
            ):
                documents += success
                logger.debug(
                    "Wrote %d documents [action: %s ok: %d]", documents, action, success
                )
        except BulkIndexError as error:
            raise BackendException(
                *error.args, f"{documents} succeeded writes"
            ) from error
        return documents

    def query_statements(self, params: StatementParameters) -> StatementQueryResult:
        """Returns the results of a statements query using xAPI parameters."""
        # pylint: disable=too-many-branches

        es_query_filters = []

        if params.statementId:
            es_query_filters += [{"term": {"_id": params.statementId}}]

        if params.agent__mbox:
            es_query_filters += [{"term": {"actor.mbox.keyword": params.agent__mbox}}]

        if params.agent__mbox_sha1sum:
            es_query_filters += [
                {"term": {"actor.mbox_sha1sum.keyword": params.agent__mbox_sha1sum}}
            ]

        if params.agent__openid:
            es_query_filters += [
                {"term": {"actor.openid.keyword": params.agent__openid}}
            ]

        if params.agent__account__name:
            es_query_filters += [
                {"term": {"actor.account.name.keyword": params.agent__account__name}},
                {
                    "term": {
                        "actor.account.homePage.keyword": params.agent__account__home_page  # pylint: disable=line-too-long # noqa: E501
                    }
                },
            ]

        if params.verb:
            es_query_filters += [{"term": {"verb.id.keyword": params.verb}}]

        if params.activity:
            es_query_filters += [
                {"term": {"object.objectType.keyword": "Activity"}},
                {"term": {"object.id.keyword": params.activity}},
            ]

        if params.since:
            es_query_filters += [{"range": {"timestamp": {"gt": params.since}}}]

        if params.until:
            es_query_filters += [{"range": {"timestamp": {"lte": params.until}}}]

        if len(es_query_filters) > 0:
            es_query = {"query": {"bool": {"filter": es_query_filters}}}
        else:
            es_query = {"query": {"match_all": {}}}

        # Honor the "ascending" parameter, otherwise show most recent statements first
        es_query.update(
            {"sort": [{"timestamp": {"order": "asc" if params.ascending else "desc"}}]}
        )

        if params.search_after:
            es_query.update({"search_after": params.search_after.split("|")})

        # Disable total hits counting for performance as we're not using it.
        es_query.update({"track_total_hits": False})

        if not params.pit_id:
            pit_response = self._open_point_in_time(
                index=self.index, keep_alive=settings.RUNSERVER_POINT_IN_TIME_KEEP_ALIVE
            )
            params.pit_id = pit_response["id"]

        es_query.update(
            {
                "pit": {
                    "id": params.pit_id,
                    # extend duration of PIT whenever it is used
                    "keep_alive": settings.RUNSERVER_POINT_IN_TIME_KEEP_ALIVE,
                }
            }
        )
        es_response = self._search(body=es_query, size=params.limit)
        es_documents = es_response["hits"]["hits"]
        search_after = None
        if es_documents:
            search_after = "|".join([str(part) for part in es_documents[-1]["sort"]])

        return StatementQueryResult(
            statements=[document["_source"] for document in es_documents],
            pit_id=es_response["pit_id"],
            search_after=search_after,
        )

    def query_statements_by_ids(self, ids: List[str]) -> List:
        """Returns the list of matching statement IDs from the database."""
        body = {"query": {"terms": {"_id": ids}}}
        return self._search(index=self.index, body=body)["hits"]["hits"]

    def _search(self, **kwargs):
        """Wraps the ElasticSearch.search method.

        Raises:
            BackendException: raised for any failure.
        """
        try:
            return self.client.search(**kwargs)
        except ApiError as error:
            msg = "Failed to execute ElasticSearch query"
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error

    def _open_point_in_time(self, **kwargs):
        """Wraps the ElasticSearch.open_point_in_time method.

        Raises:
            BackendException: raised for any failure.
        """
        try:
            return self.client.open_point_in_time(**kwargs)
        except (ApiError, ValueError) as error:
            msg = "Failed to open ElasticSearch point in time"
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error



"""API routes related to statements."""

import logging
from datetime import datetime
from typing import List, Literal, Optional, Union
from urllib.parse import ParseResult, urlencode
from uuid import UUID, uuid4

from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    HTTPException,
    Query,
    Request,
    status,
)
from pydantic import parse_raw_as
from pydantic.types import Json

from ralph.api.auth import get_authenticated_user
from ralph.api.forwarding import forward_xapi_statements, get_active_xapi_forwardings
from ralph.backends.database.base import BaseDatabase, StatementParameters
from ralph.conf import settings
from ralph.exceptions import BackendException, BadFormatException
from ralph.models.xapi.base.agents import (
    BaseXapiAgent,
    BaseXapiAgentWithAccount,
    BaseXapiAgentWithMbox,
    BaseXapiAgentWithMboxSha1Sum,
    BaseXapiAgentWithOpenId,
)

from ..models import ErrorDetail, LaxStatement

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/xAPI/statements",
    dependencies=[Depends(get_authenticated_user)],
)


DATABASE_CLIENT: BaseDatabase = getattr(
    settings.BACKENDS.DATABASE, settings.RUNSERVER_BACKEND.upper()
).get_instance()

POST_PUT_RESPONSES = {
    400: {
        "model": ErrorDetail,
        "description": "The request was invalid.",
    },
    409: {
        "model": ErrorDetail,
        "description": "Statements had a conflict with existing statements.",
    },
}


@router.get("")
@router.get("/")
# pylint: disable=too-many-arguments, too-many-locals
async def get(
    request: Request,
    ###
    # Query string parameters defined by the LRS specification
    ###
    # pylint: disable=invalid-name
    statementId: Optional[str] = Query(None, description="Id of Statement to fetch"),
    # pylint: disable=invalid-name, unused-argument
    voidedStatementId: Optional[str] = Query(
        None, description="**Not implemented** Id of voided Statement to fetch"
    ),
    agent: Optional[Json] = Query(
        None,
        description=(
            "Filter, only return Statements for which the specified "
            "Agent or Group is the Actor or Object of the Statement"
        ),
    ),
    verb: Optional[str] = Query(
        None,
        description="Filter, only return Statements matching the specified Verb id",
    ),
    activity: Optional[str] = Query(
        None,
        description=(
            "Filter, only return Statements for which the Object "
            "of the Statement is an Activity with the specified id"
        ),
    ),
    # pylint: disable=unused-argument
    registration: Optional[UUID] = Query(
        None,
        description=(
            "**Not implemented** "
            "Filter, only return Statements matching the specified registration id"
        ),
    ),
    # pylint: disable=unused-argument
    related_activities: Optional[bool] = Query(
        False,
        description=(
            "**Not implemented** "
            "Apply the Activity filter broadly. Include Statements for which "
            "the Object, any of the context Activities, or any of those properties "
            "in a contained SubStatement match the Activity parameter, "
            "instead of that parameter's normal behaviour"
        ),
    ),
    # pylint: disable=unused-argument
    related_agents: Optional[bool] = Query(
        False,
        description=(
            "**Not implemented** "
            "Apply the Agent filter broadly. Include Statements for which "
            "the Actor, Object, Authority, Instructor, Team, or any of these "
            "properties in a contained SubStatement match the Agent parameter, "
            "instead of that parameter's normal behaviour."
        ),
    ),
    since: Optional[datetime] = Query(
        None,
        description=(
            "Only Statements stored since the "
            "specified Timestamp (exclusive) are returned"
        ),
    ),
    until: Optional[datetime] = Query(
        None,
        description=(
            "Only Statements stored at or "
            "before the specified Timestamp are returned"
        ),
    ),
    limit: Optional[int] = Query(
        settings.RUNSERVER_MAX_SEARCH_HITS_COUNT,
        description=(
            "Maximum number of Statements to return. "
            "0 indicates return the maximum the server will allow"
        ),
    ),
    # pylint: disable=unused-argument, redefined-builtin
    format: Optional[Literal["ids", "exact", "canonical"]] = Query(
        "exact",
        description=(
            "**Not implemented** "
            'If "ids", only include minimum information necessary in Agent, Activity, '
            "Verb and Group Objects to identify them. For Anonymous Groups this means "
            "including the minimum information needed to identify each member. "
            'If "exact", return Agent, Activity, Verb and Group Objects populated '
            "exactly as they were when the Statement was received. An LRS requesting "
            "Statements for the purpose of importing them would use a format of "
            '"exact" in order to maintain Statement Immutability. '
            'If "canonical", return Activity Objects and Verbs populated with the '
            "canonical definition of the Activity Objects and Display of the Verbs "
            "as determined by the LRS, after applying the language filtering process "
            "defined below, and return the original Agent and Group Objects "
            'as in "exact" mode.'
        ),
    ),
    # pylint: disable=unused-argument
    attachments: Optional[bool] = Query(
        False,
        description=(
            "**Not implemented** "
            'If "true", the LRS uses the multipart response format and includes '
            'all attachments as described previously. If "false", the LRS sends '
            "the prescribed response with Content-Type application/json and "
            "does not send attachment data."
        ),
    ),
    ascending: Optional[bool] = Query(
        False, description='If "true", return results in ascending order of stored time'
    ),
    ###
    # Private use query string parameters
    ###
    search_after: Optional[str] = Query(
        None,
        description=(
            "Sorting data to allow pagination through large number of search results. "
            "NB: for internal use, not part of the LRS specification."
        ),
    ),
    pit_id: Optional[str] = Query(
        None,
        description=(
            "Point-in-time ID to ensure consistency of search requests through "
            "multiple pages."
            "NB: for internal use, not part of the LRS specification."
        ),
    ),
):
    """Fetches a single xAPI Statement or multiple xAPI Statements.

    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#213-get-statements
    """
    # Make sure the limit does not go above max from settings
    limit = min(limit, settings.RUNSERVER_MAX_SEARCH_HITS_COUNT)

    # 400 Bad Request for requests using both `statementId` and `voidedStatementId`
    if (statementId is not None) and (voidedStatementId is not None):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                "Query parameters cannot include both `statementId`"
                "and `voidedStatementId`"
            ),
        )

    # 400 Bad Request for any request containing `statementId` or
    # `voidedStatementId` and any other parameter besides `attachments` or `format`.
    # NB: `limit` and `ascending` are not handled to simplify implementation, and as it
    # has no incidence on UX, when fetching a single statement
    excluded_params = [
        agent,
        verb,
        activity,
        registration,
        related_activities,
        related_agents,
        since,
        until,
    ]
    # NB: bool(param) relies on all defaults being None, 0, or False
    if (statementId or voidedStatementId) and any(
        bool(param) for param in excluded_params
    ):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                "Querying by id only accepts `attachments` and `format` as extra"
                "parameters"
            ),
        )

    # Parse the agent parameter (JSON) into multiple string parameters
    query_params = dict(request.query_params)
    if query_params.get("agent") is not None:
        # Transform agent to `dict` as FastAPI cannot parse JSON (seen as string)
        agent = parse_raw_as(BaseXapiAgent, query_params["agent"])

        query_params.pop("agent")

        if isinstance(agent, BaseXapiAgentWithMbox):
            query_params["agent__mbox"] = agent.mbox
        elif isinstance(agent, BaseXapiAgentWithMboxSha1Sum):
            query_params["agent__mbox_sha1sum"] = agent.mbox_sha1sum
        elif isinstance(agent, BaseXapiAgentWithOpenId):
            query_params["agent__openid"] = agent.openid
        elif isinstance(agent, BaseXapiAgentWithAccount):
            query_params["agent__account__name"] = agent.account.name
            query_params["agent__account__home_page"] = agent.account.homePage

    # Query Database
    try:
        query_result = DATABASE_CLIENT.query_statements(
            StatementParameters(**{**query_params, "limit": limit})
        )
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    # Prepare the link to get the next page of the request, while preserving the
    # consistency of search results.
    # NB: There is an unhandled edge case where the total number of results is
    # exactly a multiple of the "limit", in which case we'll offer an extra page
    # with 0 results.
    response = {}
    if len(query_result.statements) == limit:
        # Search after relies on sorting info located in the last hit
        path = request.url.path
        query = dict(request.query_params)

        query.update(
            {
                "pit_id": query_result.pit_id,
                "search_after": query_result.search_after,
            }
        )

        response.update(
            {
                "more": ParseResult(
                    scheme="",
                    netloc="",
                    path=path,
                    params="",
                    query=urlencode(query),
                    fragment="",
                ).geturl(),
            }
        )

    return {**response, "statements": query_result.statements}


@router.put("/", responses=POST_PUT_RESPONSES, status_code=status.HTTP_204_NO_CONTENT)
@router.put("", responses=POST_PUT_RESPONSES, status_code=status.HTTP_204_NO_CONTENT)
# pylint: disable=unused-argument
async def put(
    # pylint: disable=invalid-name
    statementId: str,
    statement: LaxStatement,
    background_tasks: BackgroundTasks,
):
    """Stores a single statement as a single member of a set.

    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#211-put-statements
    """
    statement_dict = {statementId: statement.dict(exclude_unset=True)}

    # Force the UUID id in the statement to string, make sure it matches the
    # statementId given in the URL.
    statement_dict[statementId]["id"] = str(statement_dict[statementId]["id"])

    if not statementId == statement_dict[statementId]["id"]:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="xAPI statement id does not match given statementId",
        )

    if get_active_xapi_forwardings():
        background_tasks.add_task(
            forward_xapi_statements, list(statement_dict.values())
        )

    try:
        statements_ids_result = DATABASE_CLIENT.query_statements_by_ids([statementId])
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    if len(statements_ids_result) > 0:
        # NB: LRS specification calls for performing a deep comparison of incoming
        # statements and existing statements with the same ID.
        # This seems too costly for performance and was not implemented for this POC.
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="Statement already exists with the same ID",
        )

    # For valid requests, perform the bulk indexing of all incoming statements
    try:
        success_count = DATABASE_CLIENT.put(
            statement_dict.values(), ignore_errors=False
        )
    except (BackendException, BadFormatException) as exc:
        logger.error("Failed to index submitted statement")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Statement indexation failed",
        ) from exc

    logger.info("Indexed %d statements with success", success_count)


@router.post("/", responses=POST_PUT_RESPONSES)
@router.post("", responses=POST_PUT_RESPONSES)
async def post(
    statements: Union[LaxStatement, List[LaxStatement]],
    background_tasks: BackgroundTasks,
):
    """Stores a set of statements (or a single statement as a single member of a set).

    NB: at this time, using POST to make a GET request, is not supported.
    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#212-post-statements
    """
    # As we accept both a single statement as a dict, and multiple statements as a list,
    # we need to normalize the data into a list in all cases before we can process it.
    if not isinstance(statements, list):
        statements = [statements]

    # The statements dict has multiple functions:
    # - generate IDs for statements that are missing them;
    # - use the list of keys to perform validations and as a final return value;
    # - provide an iterable containing both the statements and generated IDs for bulk.
    statements_dict = {}
    for statement in map(lambda x: x.dict(exclude_unset=True), statements):
        statement_id = str(statement.get("id", uuid4()))
        statement["id"] = statement_id
        statements_dict[statement_id] = statement

    # Requests with duplicate statement IDs are considered invalid
    # statements_ids were deduplicated by the dict, statements list was not
    statements_ids = list(statements_dict.keys())
    if len(statements) != len(statements_ids):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Duplicate statement IDs in the list of statements",
        )

    if get_active_xapi_forwardings():
        background_tasks.add_task(
            forward_xapi_statements, list(statements_dict.values())
        )

    try:
        statements_ids_result = DATABASE_CLIENT.query_statements_by_ids(statements_ids)
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    if len(statements_ids_result) > 0:
        # NB: LRS specification calls for performing a deep comparison of incoming
        # statements and existing statements with the same ID.
        # This seems too costly for performance and was not implemented for this POC.
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="Statements already exist with the same ID",
        )

    # For valid requests, perform the bulk indexing of all incoming statements
    try:
        success_count = DATABASE_CLIENT.put(
            statements_dict.values(), ignore_errors=False
        )
    except (BackendException, BadFormatException) as exc:
        logger.error("Failed to index submitted statements")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Statements bulk indexation failed",
        ) from exc

    logger.info("Indexed %d statements with success", success_count)

    return statements_ids



"""Video xAPI events context fields definitions."""

from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from uuid import UUID

from pydantic import Field, NonNegativeFloat

from ..base.contexts import BaseXapiContext, BaseXapiContextContextActivities
from ..base.unnested_objects import BaseXapiActivity
from ..concepts.activity_types.scorm_profile import ProfileActivity
from ..concepts.constants.video import (
    CONTEXT_EXTENSION_CC_ENABLED,
    CONTEXT_EXTENSION_CC_SUBTITLE_LANG,
    CONTEXT_EXTENSION_COMPLETION_THRESHOLD,
    CONTEXT_EXTENSION_FULL_SCREEN,
    CONTEXT_EXTENSION_LENGTH,
    CONTEXT_EXTENSION_SCREEN_SIZE,
    CONTEXT_EXTENSION_SESSION_ID,
    CONTEXT_EXTENSION_SPEED,
    CONTEXT_EXTENSION_USER_AGENT,
    CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE,
    CONTEXT_EXTENSION_VOLUME,
)
from ..config import BaseExtensionModelWithConfig


class VideoContextContextActivitiesCategory(BaseXapiActivity):
    # noqa: D205, D415
    """Pydantic model for video `context`.`contextActivities`.`category`
    property.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/video`.
        definition (dict): see ProfileActivity.
    """

    id: Literal["https://w3id.org/xapi/video"] = "https://w3id.org/xapi/video"
    definition: ProfileActivity


class VideoContextContextActivities(BaseXapiContextContextActivities):
    """Pydantic model for video `context`.`contextActivities` property.

    Attributes:
        category (list): see VideoContextContextActivitiesCategory.
    """

    category: Union[
        VideoContextContextActivitiesCategory,
        List[VideoContextContextActivitiesCategory],
    ]


class BaseVideoContext(BaseXapiContext):
    """Pydantic model for video core `context` property.

    Attributes:
        contextActivities (dict): see VideoContextContextActivities.
    """

    contextActivities: Optional[VideoContextContextActivities]


class VideoContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for video core context `extensions` property.

    Attributes:
        session (uuid): Consists of the ID of the active session.
    """

    session_id: Optional[UUID] = Field(alias=CONTEXT_EXTENSION_SESSION_ID)


class VideoInitializedContextExtensions(VideoContextExtensions):
    """Pydantic model for video initialized `context` `extensions` property.

    Attributes:
        length (float): Consists of the length of the video.
        ccSubtitleEnabled (bool): Indicates whether subtitle or closed captioning is
            enabled.
        ccSubtitleLanguage (str): Consists of the language of subtitle or closed
            captioning.
        fullScreen (bool): Indicates whether the video is played in full screen mode.
        screenSize (str): Consists of the device playback screen size or the maximum
            available screen size for Video playback.
        videoPlaybackSize (str): Consists of the size in Width x Height of the video as
            viewed by the user.
        speed (str): Consists of the play back speed.
        userAgent (str): Consists of the User Agent string of the browser,
            if the video is launched in browser.
        volume (int): Consists of the volume of the video.
        completionThreshold (float): Consists of the percentage of media that should be
            consumed to trigger a completion.
    """

    length: NonNegativeFloat = Field(alias=CONTEXT_EXTENSION_LENGTH)
    ccSubtitleEnabled: Optional[bool] = Field(alias=CONTEXT_EXTENSION_CC_ENABLED)
    ccSubtitleLang: Optional[str] = Field(alias=CONTEXT_EXTENSION_CC_SUBTITLE_LANG)
    fullScreen: Optional[bool] = Field(alias=CONTEXT_EXTENSION_FULL_SCREEN)
    screenSize: Optional[str] = Field(alias=CONTEXT_EXTENSION_SCREEN_SIZE)
    videoPlaybackSize: Optional[str] = Field(
        alias=CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE
    )
    speed: Optional[str] = Field(alias=CONTEXT_EXTENSION_SPEED)
    userAgent: Optional[str] = Field(alias=CONTEXT_EXTENSION_USER_AGENT)
    volume: Optional[int] = Field(alias=CONTEXT_EXTENSION_VOLUME)
    completionThreshold: Optional[float] = Field(
        alias=CONTEXT_EXTENSION_COMPLETION_THRESHOLD
    )


class VideoBrowsingContextExtensions(VideoContextExtensions):
    """Pydantic model for video browsing `context`.`extensions` property.

    Such field is used in `paused`, `completed` and `terminated` events.

    Attributes:
        completionThreshold (float): Consists of the percentage of media that should
            be consumed to trigger a completion.
        length (float): Consists of the length of the video.
    """

    length: NonNegativeFloat = Field(alias=CONTEXT_EXTENSION_LENGTH)
    completionThreshold: Optional[float] = Field(
        alias=CONTEXT_EXTENSION_COMPLETION_THRESHOLD
    )


class VideoEnableClosedCaptioningContextExtensions(VideoContextExtensions):
    """Represents the context.extensions field for video `interacted` xAPI statement.

    Attributes:
        ccSubtitleLanguage (str): Consists of the language of subtitle or closed
            captioning.
    """

    ccSubtitleLanguage: str = Field(alias=CONTEXT_EXTENSION_CC_SUBTITLE_LANG)


class VideoVolumeChangeInteractionContextExtensions(VideoContextExtensions):
    # noqa: D205, D415
    """Pydantic model for video volume change interaction `context`.`extensions`
    property.

    Attributes:
        volume (int): Consists of the volume of the video.
    """

    volume: int = Field(alias=CONTEXT_EXTENSION_VOLUME)


class VideoScreenChangeInteractionContextExtensions(VideoContextExtensions):
    # noqa: D205, D415
    """Pydantic model for video screen change interaction `context`.`extensions`
    property.

    Attributes:
        fullScreen (bool): Indicates whether the video is played in full screen mode.
        screenSize (str): Expresses the total available screen size for Video playback.
        videoPlaybackSize (str): Consists of the size in Width x Height of the video as
            viewed by the user.
    """

    fullScreen: bool = Field(alias=CONTEXT_EXTENSION_FULL_SCREEN)
    screenSize: str = Field(alias=CONTEXT_EXTENSION_SCREEN_SIZE)
    videoPlaybackSize: str = Field(alias=CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE)


class VideoInitializedContext(BaseVideoContext):
    """Pydantic model for video initialized `context` property.

    Attributes:
        extensions (dict): See VideoInitializedContextExtensions.
    """

    extensions: VideoInitializedContextExtensions


class VideoPlayedContext(BaseVideoContext):
    """Pydantic model for video played `context` property.

    Attributes:
        extensions (dict): See VideoContextExtensions.
    """

    extensions: Optional[VideoContextExtensions]


class VideoPausedContext(BaseVideoContext):
    """Pydantic model for video paused `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoSeekedContext(BaseVideoContext):
    """Pydantic model for video seeked `context` property.

    Attributes:
        extensions (dict): See VideoContextExtensions.
    """

    extensions: Optional[VideoContextExtensions]


class VideoCompletedContext(BaseVideoContext):
    """Pydantic model for video completed `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoTerminatedContext(BaseVideoContext):
    """Pydantic model for video terminated `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoEnableClosedCaptioningContext(BaseVideoContext):
    """Pydantic model for video enable closed captioning `context` property.

    Attributes:
        extensions (dict): See VideoEnableClosedCaptioningContextExtensions.
    """

    extensions: VideoEnableClosedCaptioningContextExtensions


class VideoVolumeChangeInteractionContext(BaseVideoContext):
    """Pydantic model for video volume change interaction `context` property.

    Attributes:
        extensions (dict): See VideoVolumeChangeInteractionContextExtensions.
    """

    extensions: VideoVolumeChangeInteractionContextExtensions


class VideoScreenChangeInteractionContext(BaseVideoContext):
    """Pydantic model for video screen change interaction `context` property.

    Attributes:
        extensions (dict): See VideoScreenChangeInteractionContextExtensions.
    """

    extensions: VideoScreenChangeInteractionContextExtensions



"""API routes related to application health checking."""

import logging

from fastapi import APIRouter, status
from fastapi.responses import JSONResponse

from ralph.backends.database.base import BaseDatabase
from ralph.conf import settings

logger = logging.getLogger(__name__)

router = APIRouter()

DATABASE_CLIENT: BaseDatabase = getattr(
    settings.BACKENDS.DATABASE, settings.RUNSERVER_BACKEND.upper()
).get_instance()


@router.get("/__lbheartbeat__")
async def lbheartbeat():
    """Load balancer heartbeat.

    Returns a 200 when the server is running.
    """
    return


@router.get("/__heartbeat__")
async def heartbeat():
    """Application heartbeat.

    Returns a 200 if all checks are successful.
    """
    content = {"database": DATABASE_CLIENT.status().value}
    status_code = (
        status.HTTP_200_OK
        if all(v == "ok" for v in content.values())
        else status.HTTP_500_INTERNAL_SERVER_ERROR
    )

    return JSONResponse(status_code=status_code, content=content)



"""Problem interaction events model definitions."""

from typing import List, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from pydantic import Json

from ralph.models.selector import selector

from ..browser import BaseBrowserModel
from ..server import BaseServerModel
from .fields.events import (
    EdxProblemHintDemandhintDisplayedEventField,
    EdxProblemHintFeedbackDisplayedEventField,
    ProblemCheckEventField,
    ProblemCheckFailEventField,
    ProblemRescoreEventField,
    ProblemRescoreFailEventField,
    ResetProblemEventField,
    ResetProblemFailEventField,
    SaveProblemFailEventField,
    SaveProblemSuccessEventField,
    ShowAnswerEventField,
    UIProblemResetEventField,
    UIProblemShowEventField,
)


class EdxProblemHintDemandhintDisplayed(BaseServerModel):
    """Pydantic model for `edx.problem.hint.demandhint_displayed` statement.

    The server emits this statement when a user requests a hint for a problem.

    Attributes:
        event (dict): See EdxProblemHintDemandhintDisplayedEventField.
        event_type (str): Consists of the value `edx.problem.hint.demandhint_displayed`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.problem.hint.demandhint_displayed"
    )

    event: EdxProblemHintDemandhintDisplayedEventField
    event_type: Literal["edx.problem.hint.demandhint_displayed"]
    page: Literal["x_module"]


class EdxProblemHintFeedbackDisplayed(BaseServerModel):
    """Pydantic model for `edx.problem.hint.feedback_displayed` statement.

    The server emits this event when a user receives a hint after answering a problem.

    Attributes:
        event (dict): See EdxProblemHintFeedbackDisplayedEventField.
        event_type (str): Consists of the value `edx.problem.hint.feedback_displayed`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.problem.hint.feedback_displayed"
    )

    event: EdxProblemHintFeedbackDisplayedEventField
    event_type: Literal["edx.problem.hint.feedback_displayed"]
    page: Literal["x_module"]


class UIProblemCheck(BaseBrowserModel):
    """Pydantic model for `problem_check` statement.

    The browser emits this event when a user checks a problem.

    Attributes:
        event (str): Consists of values of problem being checked, styled as `GET`
            parameters.
        event_type (str): Consists of the value `problem_check`.
        name (str): Consists of the value `problem_check`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_check")

    event: str
    event_type: Literal["problem_check"]
    name: Literal["problem_check"]


class ProblemCheck(BaseServerModel):
    """Pydantic model for `problem_check` statement.

    The server emits this event when a user checks a problem.

    Attributes:
        event (dict): See ProblemCheckEventField.
        event_type (str): Consists of the value `problem_check`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_check")

    event: ProblemCheckEventField
    event_type: Literal["problem_check"]
    page: Literal["x_module"]


class ProblemCheckFail(BaseServerModel):
    """Pydantic model for `problem_check_fail` statement.

    The server emits this event when a user checks a problem and a failure prevents the
    problem from being checked successfully.

    Attributes:
        event (dict): See ProblemCheckFailEventField.
        event_type (str): Consists of the value `problem_check_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_check_fail")

    event: ProblemCheckFailEventField
    event_type: Literal["problem_check_fail"]
    page: Literal["x_module"]


class UIProblemGraded(BaseBrowserModel):
    """Pydantic model for `problem_graded` statement.

    The server emits this statement each time a user clicks <kbd>Check</kbd> for a
    problem and it is graded successfully.

    Attributes:
        event (list): See ProblemGradedEventField.
        event_type (str): Consists of the value `problem_graded`.
        name (str): Consists of the value `problem_graded`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_graded")

    event: List[Union[str, Literal[None], None]]
    event_type: Literal["problem_graded"]
    name: Literal["problem_graded"]


class ProblemRescore(BaseServerModel):
    """Pydantic model for `problem_rescore` statement.

    The server emits this statement when a problem is successfully rescored.

    Attributes:
        event (dict): See ProblemRescoreEventField.
        event_type (str): Consists of the value `problem_rescore`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_rescore")

    event: ProblemRescoreEventField
    event_type: Literal["problem_rescore"]
    page: Literal["x_module"]


class ProblemRescoreFail(BaseServerModel):
    """Pydantic model for `problem_rescore_fail` statement.

    The server emits this statement when a problem cannot be successfully rescored.

    Attributes:
        event (dict): See ProblemRescoreFailEventField.
        event_type (str): Consists of the value `problem_rescore_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_rescore_fail")

    event: ProblemRescoreFailEventField
    event_type: Literal["problem_rescore_fail"]
    page: Literal["x_module"]


class UIProblemReset(BaseBrowserModel):
    """Pydantic model for `problem_reset` statement.

    The browser emits problem_reset events when a user clicks <kbd>Reset</kbd> to reset
    the problem answer.

    Attributes:
        event (json): See ProblemResetEventField.
        event_type (str): Consists of the value `problem_reset`.
        name (str): Consists of the value `problem_reset`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_reset")

    event: Union[
        str,
        Json[UIProblemResetEventField],  # pylint: disable=unsubscriptable-object
        UIProblemResetEventField,
    ]
    event_type: Literal["problem_reset"]
    name: Literal["problem_reset"]


class UIProblemSave(BaseBrowserModel):
    """Pydantic model for `problem_save` statement.

    The browser emits this statement when a user saves a problem.

    Attributes:
        event (str): Consists of all the answers saved for the problem.
        event_type (str): Consists of the value `problem_save`.
        name (str): Consists of the value `problem_save`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_save")

    event: str
    event_type: Literal["problem_save"]
    name: Literal["problem_save"]


class UIProblemShow(BaseBrowserModel):
    """Pydantic model for `problem_show` statement.

    The browser emits this statement when the answer clicks <kbd>Show Answer</kbd> to
    show the problem answer.

    Attributes:
        event (json): See ProblemShowEventField.
        event_type (str): Consists of the value `problem_save`.
        name (str): Consists of the value `problem_save`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_show")

    event: Union[
        Json[UIProblemShowEventField],  # pylint: disable=unsubscriptable-object
        UIProblemShowEventField,
    ]
    event_type: Literal["problem_show"]
    name: Literal["problem_show"]


class ResetProblem(BaseServerModel):
    """Pydantic model for `reset_problem` statement.

    The server emits this statement when a problem has been reset successfully.

    Attributes:
        event (dict): See ResetProblemEventField.
        event_type (str): Consists of the value `reset_problem`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="reset_problem")

    event: ResetProblemEventField
    event_type: Literal["reset_problem"]
    page: Literal["x_module"]


class ResetProblemFail(BaseServerModel):
    """Pydantic model for `reset_problem_fail` statement.

    The server emits this statement when a problem cannot be reset successfully.

    Attributes:
        event (dict): See ResetProblemFailEventField.
        event_type (str): Consists of the value `reset_problem_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="reset_problem_fail")

    event: ResetProblemFailEventField
    event_type: Literal["reset_problem_fail"]
    page: Literal["x_module"]


class SaveProblemFail(BaseServerModel):
    """Pydantic model for `save_problem_fail` statement.

    The server emits this statement when a problem cannot be saved successfully.

    Attributes:
        event (dict): See SaveProblemFailEventField.
        event_type (str): Consists of the value `save_problem_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="save_problem_fail")

    event: SaveProblemFailEventField
    event_type: Literal["save_problem_fail"]
    page: Literal["x_module"]


class SaveProblemSuccess(BaseServerModel):
    """Pydantic model for `save_problem_success` statement.

    The server emits this statement when a problem is saved successfully.

    Attributes:
        event (dict): See SaveProblemSuccessEventField.
        event_type (str): Consists of the value `save_problem_success`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="save_problem_success")

    event: SaveProblemSuccessEventField
    event_type: Literal["save_problem_success"]
    page: Literal["x_module"]


class ShowAnswer(BaseServerModel):
    """Pydantic model for `showanswer` statement.

    The server emits this statement when the answer to a problem is shown.

    Attributes:
        event (dict): See ShowAnswerEventField.
        event_type (str): Consists of the value `showanswer`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="showanswer")

    event: ShowAnswerEventField
    event_type: Literal["showanswer"]
    page: Literal["x_module"]



"""Constants for `TinCan Vocabulary` xAPI profile."""

# Context extension IRIs
CONTEXT_EXTENSION_PLANNED_DURATION = (  # pylint:disable=invalid-name
    "http://id.tincanapi.com/extension/planned-duration"
)



"""ClickHouse database backend for Ralph."""
import datetime
import json
import logging
import uuid
from dataclasses import asdict
from typing import Generator, List, Optional, TextIO, Union

import clickhouse_connect
from clickhouse_connect.driver.exceptions import ClickHouseError
from pydantic import BaseModel, ValidationError

from ralph.conf import ClickhouseClientOptions, settings
from ralph.exceptions import BackendException, BadFormatException

from .base import (
    BaseDatabase,
    BaseQuery,
    DatabaseStatus,
    StatementParameters,
    StatementQueryResult,
    enforce_query_checks,
)

clickhouse_settings = settings.BACKENDS.DATABASE.CLICKHOUSE
logger = logging.getLogger(__name__)


class ClickHouseInsert(BaseModel):
    """Model to validate required fields for ClickHouse insertion."""

    event_id: uuid.UUID
    emission_time: datetime.datetime


class ClickHouseQuery(BaseQuery):
    """ClickHouse query model."""

    where_clause: Optional[str]
    return_fields: Optional[List[str]]


class ClickHouseDatabase(BaseDatabase):  # pylint: disable=too-many-instance-attributes
    """ClickHouse database backend."""

    name = "clickhouse"
    query_model = ClickHouseQuery

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str = clickhouse_settings.HOST,
        port: int = clickhouse_settings.PORT,
        database: str = clickhouse_settings.DATABASE,
        event_table_name: str = clickhouse_settings.EVENT_TABLE_NAME,
        username: str = clickhouse_settings.USERNAME,
        password: str = clickhouse_settings.PASSWORD,
        client_options: ClickhouseClientOptions = clickhouse_settings.CLIENT_OPTIONS,
    ):
        """Instantiates the ClickHouse configuration.

        Args:
            host (str): ClickHouse server host to connect to.
            port (int): ClickHouse server port to connect to.
            database (str): ClickHouse database to connect to.
            event_table_name (str): Table where events live.
            username (str): ClickHouse username to connect as (optional).
            password (str): Password for the given ClickHouse username (optional).
            client_options (dict): A dictionary of valid options for the ClickHouse
                client connection.

        If username and password are None, we will try to connect as the ClickHouse
        user "default".
        """
        if client_options is None:
            client_options = {
                "date_time_input_format": "best_effort",  # Allows RFC dates
                "allow_experimental_object_type": 1,  # Allows JSON data type
            }
        else:
            client_options = client_options.dict()

        self.host = host
        self.port = port
        self.database = database
        self.event_table_name = event_table_name
        self.username = username
        self.password = password
        self.client_options = client_options
        self._client = None

    @property
    def client(self):
        """Creates a ClickHouse client if it doesn't exist.

        We do this here so that we don't interrupt initialization in the case
        where ClickHouse is not running when Ralph starts up, which will cause
        Ralph to hang. This client is HTTP, so not actually stateful. Ralph
        should be able to gracefully deal with ClickHouse outages at all other
        times.
        """
        if not self._client:
            self._client = clickhouse_connect.get_client(
                host=self.host,
                port=self.port,
                database=self.database,
                username=self.username,
                password=self.password,
                settings=self.client_options,
            )
        return self._client

    def status(self) -> DatabaseStatus:
        """Checks ClickHouse connection status."""
        try:
            self.client.query("SELECT 1")
        except ClickHouseError:
            return DatabaseStatus.AWAY

        return DatabaseStatus.OK

    @enforce_query_checks
    def get(self, query: ClickHouseQuery = None, chunk_size: int = 500):
        """Gets table rows and yields them."""
        fields = ",".join(query.return_fields) if query.return_fields else "event"

        sql = f"SELECT {fields} FROM {self.event_table_name}"  # nosec

        if query.where_clause:
            sql += f"  WHERE {query.where_clause}"

        result = self.client.query(sql).named_results()

        for statement in result:
            yield statement

    @staticmethod
    def to_documents(
        stream: Union[TextIO, List], ignore_errors: bool = False
    ) -> Generator[dict, None, None]:
        """Converts `stream` lines (one statement per line) to insert tuples."""
        for line in stream:
            statement = json.loads(line) if isinstance(line, str) else line

            try:
                insert = ClickHouseInsert(
                    event_id=statement["id"], emission_time=statement["timestamp"]
                )
            except (KeyError, ValidationError) as exc:
                err = (
                    "Statement has an invalid or missing id or "
                    f"timestamp field: {statement}"
                )
                if ignore_errors:
                    logger.warning(err)
                    continue
                raise BadFormatException(err) from exc

            document = (
                insert.event_id,
                insert.emission_time,
                statement,
                json.dumps(statement),
            )

            yield document

    def bulk_import(self, batch: List, ignore_errors: bool = False) -> int:
        """Inserts a batch of documents into the selected database table."""
        try:
            # ClickHouse does not do unique keys. This is a "best effort" to
            # at least check for duplicates in each batch. Overall ID checking
            # against the database happens upstream in the POST / PUT methods.
            #
            # As opposed to Mongo, the entire batch is guaranteed to fail here
            # if any dupes are found.
            found_ids = {x[0] for x in batch}

            if len(found_ids) != len(batch):
                raise BackendException("Duplicate IDs found in batch")

            self.client.insert(
                self.event_table_name,
                batch,
                column_names=[
                    "event_id",
                    "emission_time",
                    "event",
                    "event_str",
                ],
                # Allow ClickHouse to buffer the insert, and wait for the
                # buffer to flush. Should be configurable, but I think these are
                # reasonable defaults.
                settings={"async_insert": 1, "wait_for_async_insert": 1},
            )
        except (ClickHouseError, BackendException) as error:
            if not ignore_errors:
                raise BackendException(*error.args) from error
            logger.warning(
                "Bulk import failed for current chunk but you choose to ignore it.",
            )
            # There is no current way of knowing how many rows from the batch
            # succeeded, we assume 0 here.
            return 0

        logger.debug("Inserted %s documents chunk with success", len(batch))

        return len(batch)

    def put(
        self,
        stream: Union[TextIO, List],
        chunk_size: int = 500,
        ignore_errors: bool = False,
    ) -> int:
        """Writes documents from the `stream` to the instance table."""
        logger.debug(
            "Start writing to the %s table of the %s database (chunk size: %d)",
            self.event_table_name,
            self.database,
            chunk_size,
        )

        rows_inserted = 0
        batch = []
        for document in self.to_documents(stream, ignore_errors=ignore_errors):
            batch.append(document)
            if len(batch) < chunk_size:
                continue

            rows_inserted += self.bulk_import(batch, ignore_errors=ignore_errors)
            batch = []

        # Catch any remaining documents when the last batch is smaller than chunk_size
        if len(batch) > 0:
            rows_inserted += self.bulk_import(batch, ignore_errors=ignore_errors)

        logger.debug("Inserted a total of %s documents with success", rows_inserted)

        return rows_inserted

    def query_statements_by_ids(self, ids: List[str]) -> List:
        """Returns the list of matching statement IDs from the database."""

        def chunk_id_list(chunk_size=10000):
            for i in range(0, len(ids), chunk_size):
                yield ids[i : i + chunk_size]

        sql = """
                SELECT event_id
                FROM {table_name:Identifier}
                WHERE event_id IN ({ids:Array(String)})
        """

        query_context = self.client.create_query_context(
            query=sql,
            parameters={"ids": ["1"], "table_name": self.event_table_name},
            column_oriented=True,
        )

        found_ids = []

        try:
            for chunk_ids in chunk_id_list():
                query_context.set_parameter("ids", chunk_ids)
                result = self.client.query(context=query_context).named_results()
                found_ids.extend(result)

            return found_ids
        except (ClickHouseError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to execute ClickHouse query"
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error

    def query_statements(self, params: StatementParameters) -> StatementQueryResult:
        """Returns the results of a statements query using xAPI parameters."""
        params = asdict(params)
        where_clauses = []

        if params["statementId"]:
            where_clauses.append("event_id = {statementId:UUID}")

        if params["agent__mbox"]:
            where_clauses.append("event.actor.mbox = {agent__mbox:String}")

        if params["agent__mbox_sha1sum"]:
            where_clauses.append(
                "event.actor.mbox_sha1sum = {agent__mbox_sha1sum:String}"
            )

        if params["agent__openid"]:
            where_clauses.append("event.actor.openid = {agent__openid:String}")

        if params["agent__account__name"]:
            where_clauses.append(
                "event.actor.account.name = {agent__account__name:String}"
            )
            where_clauses.append(
                "event.actor.account.homePage = {agent__account__home_page:String}"
            )

        if params["verb"]:
            where_clauses.append("event.verb.id = {verb:String}")

        if params["activity"]:
            where_clauses.append("event.object.objectType = 'Activity'")
            where_clauses.append("event.object.id = {activity:String}")

        if params["since"]:
            where_clauses.append("emission_time > {since:DateTime64(6)}")

        if params["until"]:
            where_clauses.append("emission_time <= {until:DateTime64(6)}")

        if params["search_after"]:
            search_order = ">" if params["ascending"] else "<"

            where_clauses.append(
                f"(emission_time {search_order} "
                "{search_after:DateTime64(6)}"
                " OR "
                "(emission_time = {search_after:DateTime64(6)}"
                " AND "
                f"event_id {search_order} "
                "{pit_id:UUID}"
                "))"
            )

        sort_order = "ASCENDING" if params["ascending"] else "DESCENDING"
        order_by = f"emission_time {sort_order}, event_id {sort_order}"

        response = self._find(
            where=where_clauses, parameters=params, limit=params["limit"], sort=order_by
        )
        response = list(response)

        new_search_after = None
        new_pit_id = None

        if response:
            # Our search after string is a combination of event timestamp and
            # event id, so that we can avoid losing events when they have the
            # same timestamp, and also avoid sending the same event twice.
            new_search_after = response[-1]["emission_time"].isoformat()
            new_pit_id = str(response[-1]["event_id"])

        return StatementQueryResult(
            statements=[document["event"] for document in response],
            search_after=new_search_after,
            pit_id=new_pit_id,
        )

    def _find(
        self, parameters: dict, where: List = None, limit: int = None, sort: str = None
    ):
        """Wraps the ClickHouse query method.

        Raises:
            BackendException: raised for any failure.
        """
        sql = """
        SELECT event_id, emission_time, event
        FROM {event_table_name:Identifier}
        """
        if where:
            filter_str = "WHERE 1=1 AND "
            filter_str += """
            AND
            """.join(
                where
            )
            sql += filter_str
        if sort:
            sql += f"\nORDER BY {sort}"

        if limit:
            sql += f"\nLIMIT {limit}"

        parameters["event_table_name"] = self.event_table_name

        try:
            return self.client.query(sql, parameters=parameters).named_results()
        except (ClickHouseError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to execute ClickHouse query"
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error



"""Tests for Ralph cli usage strings."""
import logging

from click.testing import CliRunner

from ralph.cli import cli

test_logger = logging.getLogger("ralph")


def test_cli_auth_command_usage():
    """Test ralph auth command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["auth", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -u, --username TEXT  The user for which we generate credentials.  "
        "[required]\n"
        "  -p, --password TEXT  The password to encrypt for this user. Will be "
        "prompted\n"
        "                       if missing.  [required]\n"
        "  -s, --scope TEXT     The user scope(s). This option can be provided "
        "multiple\n"
        "                       times.  [required]\n"
        "  -w, --write          Write new credentials to the LRS authentication file.\n"
    ) in result.output

    result = runner.invoke(cli, ["auth"])
    assert result.exit_code > 0
    assert "Error: Missing option '-u' / '--username'." in result.output


def test_cli_extract_command_usage():
    """Tests ralph extract command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["extract", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -p, --parser [gelf|es]  Container format parser used to extract events\n"
        "                          [required]\n"
    ) in result.output

    result = runner.invoke(cli, ["extract"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-p' / '--parser'. Choose from:\n\tgelf,\n\tes\n"
    ) in result.output


def test_cli_validate_command_usage():
    """Tests ralph validate command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["validate", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -f, --format [edx|xapi]  Input events format to validate  [required]\n"
        "  -I, --ignore-errors      Continue validating regardless of raised errors\n"
        "  -F, --fail-on-unknown    Stop validating at first unknown event\n"
    ) in result.output

    result = runner.invoke(cli, ["validate"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-f' / '--format'. Choose from:\n\tedx,\n\txapi\n"
    ) in result.output


def test_cli_convert_command_usage():
    """Tests ralph convert command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["convert", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  From edX to xAPI converter options: \n"
        "    -u, --uuid-namespace TEXT     The UUID namespace to use for the `ID` "
        "field\n"
        "                                  generation\n"
        "    -p, --platform-url TEXT       The `actor.account.homePage` to use in the\n"
        "                                  xAPI statements  [required]\n"
        "  -f, --from [edx]                Input events format to convert  [required]\n"
        "  -t, --to [xapi]                 Output events format  [required]\n"
        "  -I, --ignore-errors             Continue writing regardless of raised "
        "errors\n"
        "  -F, --fail-on-unknown           Stop converting at first unknown event\n"
    ) in result.output

    result = runner.invoke(cli, ["convert"])
    assert result.exit_code > 0
    assert "Error: Missing option '-p' / '--platform-url'" in result.output


def test_cli_fetch_command_usage():
    """Tests ralph fetch command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["fetch", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -b, --backend [es|mongo|clickhouse|lrs|ldp|fs|swift|s3|ws]\n"
        "                                  Backend  [required]\n"
        "  ws backend: \n"
        "    --ws-uri TEXT\n"
        "  s3 backend: \n"
        "    --s3-endpoint-url TEXT\n"
        "    --s3-bucket-name TEXT\n"
        "    --s3-default-region TEXT\n"
        "    --s3-session-token TEXT\n"
        "    --s3-secret-access-key TEXT\n"
        "    --s3-access-key-id TEXT\n"
        "  swift backend: \n"
        "    --swift-os-identity-api-version TEXT\n"
        "    --swift-os-auth-url TEXT\n"
        "    --swift-os-project-domain-name TEXT\n"
        "    --swift-os-user-domain-name TEXT\n"
        "    --swift-os-storage-url TEXT\n"
        "    --swift-os-region-name TEXT\n"
        "    --swift-os-password TEXT\n"
        "    --swift-os-username TEXT\n"
        "    --swift-os-tenant-name TEXT\n"
        "    --swift-os-tenant-id TEXT\n"
        "  fs backend: \n"
        "    --fs-path TEXT\n"
        "  ldp backend: \n"
        "    --ldp-stream-id TEXT\n"
        "    --ldp-service-name TEXT\n"
        "    --ldp-consumer-key TEXT\n"
        "    --ldp-application-secret TEXT\n"
        "    --ldp-application-key TEXT\n"
        "    --ldp-endpoint TEXT\n"
        "  lrs backend: \n"
        "    --lrs-statements-endpoint TEXT\n"
        "    --lrs-status-endpoint TEXT\n"
        "    --lrs-headers KEY=VALUE,KEY=VALUE\n"
        "    --lrs-password TEXT\n"
        "    --lrs-username TEXT\n"
        "    --lrs-base-url TEXT\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-host TEXT\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-collection TEXT\n"
        "    --mongo-database TEXT\n"
        "    --mongo-connection-uri TEXT\n"
        "  es backend: \n"
        "    --es-op-type TEXT\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "  -c, --chunk-size INTEGER        Get events by chunks of size #\n"
        "  -t, --target TEXT               Endpoint from which to fetch events (e.g.\n"
        "                                  `/statements`)\n"
        '  -q, --query \'{"KEY": "VALUE", "KEY": "VALUE"}\'\n'
        "                                  Query object as a JSON string (database\n"
        "                                  backends ONLY)\n"
    ) in result.output
    logging.warning(result.output)
    result = runner.invoke(cli, ["fetch"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-b' / '--backend'. "
        "Choose from:\n\tes,\n\tmongo,\n\tclickhouse,\n\tlrs,\n\tldp,\n\tfs,\n\tswift,"
        "\n\ts3,\n\tws\n"
    ) in result.output


def test_cli_list_command_usage():
    """Tests ralph list command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["list", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -b, --backend [ldp|fs|swift|s3]\n"
        "                                  Backend  [required]\n"
        "  s3 backend: \n"
        "    --s3-endpoint-url TEXT\n"
        "    --s3-bucket-name TEXT\n"
        "    --s3-default-region TEXT\n"
        "    --s3-session-token TEXT\n"
        "    --s3-secret-access-key TEXT\n"
        "    --s3-access-key-id TEXT\n"
        "  swift backend: \n"
        "    --swift-os-identity-api-version TEXT\n"
        "    --swift-os-auth-url TEXT\n"
        "    --swift-os-project-domain-name TEXT\n"
        "    --swift-os-user-domain-name TEXT\n"
        "    --swift-os-storage-url TEXT\n"
        "    --swift-os-region-name TEXT\n"
        "    --swift-os-password TEXT\n"
        "    --swift-os-username TEXT\n"
        "    --swift-os-tenant-name TEXT\n"
        "    --swift-os-tenant-id TEXT\n"
        "  fs backend: \n"
        "    --fs-path TEXT\n"
        "  ldp backend: \n"
        "    --ldp-stream-id TEXT\n"
        "    --ldp-service-name TEXT\n"
        "    --ldp-consumer-key TEXT\n"
        "    --ldp-application-secret TEXT\n"
        "    --ldp-application-key TEXT\n"
        "    --ldp-endpoint TEXT\n"
        "  -n, --new / -a, --all           List not fetched (or all) archives\n"
        "  -D, --details / -I, --ids       Get archives detailed output (JSON)\n"
    ) in result.output

    result = runner.invoke(cli, ["list"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-b' / '--backend'. Choose from:\n\tldp,\n\tfs,\n\t"
        "swift,\n\ts3\n"
    ) in result.output


def test_cli_runserver_command_usage():
    """Tests ralph runserver command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["runserver", "--help"])

    expected_output = (
        "Options:\n"
        "  -b, --backend [es|mongo|clickhouse]\n"
        "                                  Backend  [required]\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-host TEXT\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-collection TEXT\n"
        "    --mongo-database TEXT\n"
        "    --mongo-connection-uri TEXT\n"
        "  es backend: \n"
        "    --es-op-type TEXT\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "  -h, --host TEXT                 LRS server host name\n"
        "  -p, --port INTEGER              LRS server port\n"
    )

    assert result.exit_code == 0
    assert expected_output in result.output



"""Base event model definitions."""

from datetime import datetime
from ipaddress import IPv4Address
from pathlib import Path
from typing import Dict, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from pydantic import AnyHttpUrl, BaseModel, constr


class BaseModelWithConfig(BaseModel):
    """Pydantic model for base configuration shared among all models."""

    class Config:  # pylint: disable=missing-class-docstring # noqa: D106
        extra = "forbid"


class ContextModuleField(BaseModelWithConfig):
    """Pydantic model for `context`.`module` field.

    Attributes:
        usage_key (str): Consists of a block ID of the current component.
        display_name (str): Consists of a short description or title of the component.
    """

    usage_key: constr(regex=r"^block-v1:.+\+.+\+.+type@.+@[a-f0-9]{32}$")  # noqa:F722
    display_name: str
    original_usage_key: Optional[
        constr(
            regex=r"^block-v1:.+\+.+\+.+type@problem\+block@[a-f0-9]{32}$"  # noqa:F722
        )
    ]
    original_usage_version: Optional[str]


class BaseContextField(BaseModelWithConfig):
    """Pydantic model for core `context` field.

    Attributes:
        course_user_tags (Dict of str): Content from `user_api_usercoursetag` table.
            Retrieved with:
                `dict(
                    UserCourseTag.objects.filter(
                        user=request.user.pk, course_id=course_key
                    ).values_list('key', 'value')
                )`
            Note:
                Is only present when a course page is requested.
                Is an empty dictionary when the user is not logged in or not found in
                the `user_api_usercoursetag` table.
        user_id (int or str or None): Consists of the ID of the authenticated user.
            Retrieved with:
                `request.user.pk` querying the `auth_user` table.
            Note:
                Is an integer when the user is found in the `auth_user` table.
                Is an empty string when an exception is raised while retrieving the ID.
                Is `None` when the user is not logged in.
        org_id (str): Consists of the organization name that lists the course.
            Retrieved with:
                `course_id.org` where `course_id` is an
                `opaque_keys.edx.locator.CourseLocator` which is created using the URL
                of the requested page.
            Note:
                Is an empty string when the requested page is not a course page.
        course_id (str): Consists of the unique identifier for the visited course page.
            Retrieved with:
                `course_id.to_deprecated_string()` where `course_id` is an
                `opaque_keys.edx.locator.CourseLocator` which is created using the URL
                of the requested page.
            Note:
                Is an empty string when the requested page is not a course page.
        path (Path): Consist of the relative URL (without the hostname) of the
            requested page.
            Retrieved with:
                `request.META['PATH_INFO']`
    """

    course_id: constr(regex=r"^$|^course-v1:.+\+.+\+.+$")  # noqa:F722
    course_user_tags: Optional[Dict[str, str]]
    module: Optional[ContextModuleField]
    org_id: str
    path: Path
    user_id: Union[int, Literal[""], None]


class AbstractBaseEventField(BaseModelWithConfig):
    """Pydantic model for core `event` fields.

    The base model does not have any attributes as event field does not have common
    sub-fields.
    """


class BaseEdxModel(BaseModelWithConfig):
    """Pydantic model for core statement.

    WARNING: it does not define the `event`, `event_type` and `event_source` fields.

    Attributes:
        username (str): Consists of the unique username identifying the logged in user.
            Retrieved with:
                `request.user.username` querying the `auth_user` table.
            Note:
                Is an empty string when the user is not logged in.
                If an exception is raised when retrieving the username from the table
                then the value is `anonymous`.
                Usernames are made of 2-30 ASCII letters / numbers / underscores (_) /
                hyphens (-)
        ip (IPv4Address or str): Consists of the public IPv4 address of the user.
            Retrieved with:
                `get_ip(request)` cf. https://github.com/un33k/django-ipware/tree/1.1.0
            Note:
                Can be an empty string if the IP address is not found.
        agent (str): Consists of the `User-Agent` HTTP request header.
            Retrieved with:
                `request.META[HTTP_USER_AGENT]`
            Note:
                Can be an empty string if the header is not present in the request.
                Contains information about:
                    Browser name and version
                    Operating system name and version
                    Default language
        host (str): Consists of the hostname of the server.
            Retrieved with:
                `request.META[SERVER_NAME]`
        referer (Path): Consists of the `Referer` HTTP request header.
            Retrieved with:
                `request.META[HTTP_REFERER]`
            Note:
                Can be an empty string if the header is not present in the request.
                Contains the referring URL (previous URL visited by the user).
        accept_language (str): Consists of the `Accept-Language` HTTP request header.
            Retrieved with:
                `request.META[HTTP_ACCEPT_LANGUAGE]`
            Note:
                Can be an empty string if the header is not present in the request.
                Contains the default language settings of the user.
        context (BaseContextField): see BaseContextField.
        time (datetime): Consists of the UTC time in ISO format at which the event was
            emitted.
            Retrieved with:
                `datetime.datetime.utcnow()`
        page (None): Consists of the value `None`
            Note:
                In JSON the value is `null` instead of `None`.
    """

    username: Union[constr(min_length=2, max_length=30), Literal[""]]
    ip: Union[IPv4Address, Literal[""]]
    agent: str
    host: str
    referer: Union[AnyHttpUrl, Literal[""]]
    accept_language: str
    context: BaseContextField
    time: datetime
    page: None



"""MongoDB database backend for Ralph."""

import hashlib
import json
import logging
import struct
from typing import Generator, List, Optional, TextIO, Union

from bson.objectid import ObjectId
from dateutil.parser import isoparse
from pymongo import ASCENDING, DESCENDING, MongoClient
from pymongo.errors import BulkWriteError, ConnectionFailure, PyMongoError

from ralph.conf import MongoClientOptions, settings
from ralph.exceptions import BackendException, BadFormatException

from .base import (
    BaseDatabase,
    BaseQuery,
    DatabaseStatus,
    StatementParameters,
    StatementQueryResult,
    enforce_query_checks,
)

mongo_settings = settings.BACKENDS.DATABASE.MONGO
logger = logging.getLogger(__name__)


class MongoQuery(BaseQuery):
    """Mongo query model."""

    filter: Optional[dict]
    projection: Optional[dict]


class MongoDatabase(BaseDatabase):
    """Mongo database backend."""

    name = "mongo"
    query_model = MongoQuery

    def __init__(
        self,
        connection_uri: str = mongo_settings.CONNECTION_URI,
        database: str = mongo_settings.DATABASE,
        collection: str = mongo_settings.COLLECTION,
        client_options: MongoClientOptions = mongo_settings.CLIENT_OPTIONS,
    ):
        """Instantiates the Mongo client.

        Args:
            connection_uri (str): MongoDB connection URI.
            database (str): MongoDB database to connect to.
            collection (str): MongoDB database collection to get objects from.
            client_options (MongoClientOptions): A dictionary of valid options
            for the MongoClient class initialization.
        """
        self.client = MongoClient(connection_uri, **client_options.dict())
        self.database = getattr(self.client, database)
        self.collection = getattr(self.database, collection)

    def status(self) -> DatabaseStatus:
        """Checks MongoDB cluster connection status."""
        # Check Mongo cluster connection
        try:
            self.client.admin.command("ping")
        except ConnectionFailure:
            return DatabaseStatus.AWAY

        # Check cluster status
        if self.client.admin.command("serverStatus").get("ok", 0.0) < 1.0:
            return DatabaseStatus.ERROR

        return DatabaseStatus.OK

    @enforce_query_checks
    def get(self, query: MongoQuery = None, chunk_size: int = 500):
        """Gets collection documents and yields them.

        The `query` dictionary should only contain kwargs compatible with the
        pymongo.collection.Collection.find method signature (API reference
        documentation: https://pymongo.readthedocs.io/en/stable/api/pymongo/).
        """
        for document in self.collection.find(batch_size=chunk_size, **query.dict()):
            # Make the document json-serializable
            document.update({"_id": str(document.get("_id"))})
            yield document

    @staticmethod
    def to_documents(
        stream: Union[TextIO, list], ignore_errors: bool = False
    ) -> Generator[dict, None, None]:
        """Converts `stream` lines (one statement per line) to Mongo documents.

        We expect statements to have at least an `id` and a `timestamp` field that will
        be used to compute a unique MongoDB Object ID. This ensures that we will not
        duplicate statements in our database and allows us to support pagination.
        """
        for line in stream:
            statement = json.loads(line) if isinstance(line, str) else line
            if "id" not in statement:
                msg = f"statement {statement} has no 'id' field"
                if ignore_errors:
                    logger.warning(msg)
                    continue
                raise BadFormatException(msg)
            if "timestamp" not in statement:
                msg = f"statement {statement} has no 'timestamp' field"
                if ignore_errors:
                    logger.warning(msg)
                    continue
                raise BadFormatException(msg)
            try:
                timestamp = int(isoparse(statement["timestamp"]).timestamp())
            except ValueError as err:
                msg = f"statement {statement} has an invalid 'timestamp' field"
                if ignore_errors:
                    logger.warning(msg)
                    continue
                raise BadFormatException(msg) from err
            document = {
                "_id": ObjectId(
                    # This might become a problem in February 2106.
                    # Meanwhile, we use the timestamp in the _id field for pagination.
                    struct.pack(">I", timestamp)
                    + bytes.fromhex(
                        hashlib.sha256(bytes(statement["id"], "utf-8")).hexdigest()[:16]
                    )
                ),
                "_source": statement,
            }

            yield document

    def bulk_import(self, batch: list, ignore_errors: bool = False):
        """Inserts a batch of documents into the selected database collection."""
        try:
            new_documents = self.collection.insert_many(batch)
        except BulkWriteError as error:
            if not ignore_errors:
                raise BackendException(
                    *error.args, f"{error.details['nInserted']} succeeded writes"
                ) from error
            logger.warning(
                "Bulk importation failed for current documents chunk but you choose "
                "to ignore it.",
            )
            return error.details["nInserted"]

        inserted_count = len(new_documents.inserted_ids)
        logger.debug("Inserted %d documents chunk with success", inserted_count)

        return inserted_count

    def put(
        self,
        stream: Union[TextIO, list],
        chunk_size: int = 500,
        ignore_errors: bool = False,
    ) -> int:
        """Writes documents from the `stream` to the instance collection."""
        logger.debug(
            "Start writing to the %s collection of the %s database (chunk size: %d)",
            self.collection,
            self.database,
            chunk_size,
        )

        success = 0
        batch = []
        for document in self.to_documents(stream, ignore_errors=ignore_errors):
            batch.append(document)
            if len(batch) < chunk_size:
                continue

            success += self.bulk_import(batch, ignore_errors=ignore_errors)
            batch = []

        # Edge case: if the total number of documents is lower than the chunk size
        if len(batch) > 0:
            success += self.bulk_import(batch, ignore_errors=ignore_errors)

        logger.debug("Inserted a total of %d documents with success", success)

        return success

    def query_statements(self, params: StatementParameters) -> StatementQueryResult:
        """Returns the results of a statements query using xAPI parameters."""
        mongo_query_filters = {}

        if params.statementId:
            mongo_query_filters.update({"_source.id": params.statementId})

        if params.agent__mbox:
            mongo_query_filters.update({"_source.actor.mbox": params.agent__mbox})

        if params.agent__mbox_sha1sum:
            mongo_query_filters.update(
                {"_source.actor.mbox_sha1sum": params.agent__mbox_sha1sum}
            )

        if params.agent__openid:
            mongo_query_filters.update({"_source.actor.openid": params.agent__openid})

        if params.agent__account__name:
            mongo_query_filters.update(
                {"_source.actor.account.name": params.agent__account__name}
            )
            mongo_query_filters.update(
                {"_source.actor.account.homePage": params.agent__account__home_page}
            )

        if params.verb:
            mongo_query_filters.update({"_source.verb.id": params.verb})

        if params.activity:
            mongo_query_filters.update(
                {
                    "_source.object.objectType": "Activity",
                    "_source.object.id": params.activity,
                },
            )

        if params.since:
            mongo_query_filters.update({"_source.timestamp": {"$gt": params.since}})

        if params.until:
            mongo_query_filters.update({"_source.timestamp": {"$lte": params.until}})

        if params.search_after:
            search_order = "$gt" if params.ascending else "$lt"
            mongo_query_filters.update(
                {"_id": {search_order: ObjectId(params.search_after)}}
            )

        mongo_sort_order = ASCENDING if params.ascending else DESCENDING
        mongo_query_sort = [
            ("_source.timestamp", mongo_sort_order),
            ("_id", mongo_sort_order),
        ]

        mongo_response = self._find(
            filter=mongo_query_filters, limit=params.limit, sort=mongo_query_sort
        )
        search_after = None
        if mongo_response:
            search_after = mongo_response[-1]["_id"]

        return StatementQueryResult(
            statements=[document["_source"] for document in mongo_response],
            pit_id=None,
            search_after=search_after,
        )

    def query_statements_by_ids(self, ids: List[str]) -> List:
        """Returns the list of matching statement IDs from the database."""
        return self._find(filter={"_source.id": {"$in": ids}})

    def _find(self, **kwargs):
        """Wraps the MongoClient.collection.find method.

        Raises:
            BackendException: raised for any failure.
        """
        try:
            return list(self.collection.find(**kwargs))
        except (PyMongoError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to execute MongoDB query"
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error
